#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import functools

from django.http import HttpResponse

from eoxserver.services.auth.base import getPDP
from eoxserver.services.auth.exceptions import AuthorisationException
from eoxserver.services.ows.common.v20.encoders import OWS20ExceptionXMLEncoder


class PDPMiddleware(object):
    """ Middleware to allow authorization agains a Policy Decision Point. This
        middleware will be used for *all* requests and *all* configured views.
        If you only want to provide PDP authorization for a single view, use the
        `pdp_protect`.
    """

    def process_view(self, request, view_func, view_args, view_kwargs):
        pdp = getPDP()
        if pdp:
            try:
                authorized = pdp.authorize(request)
                message = "Not authorized"
                code = "AccessForbidden"
            except AuthorisationException as e:
                authorized = False
                message = str(e)
                code = e.code

            if not authorized:
                encoder = OWS20ExceptionXMLEncoder()
                return HttpResponse(
                    encoder.serialize(
                        encoder.encode_exception(message, "2.0.0", code)
                    ),
                    encoder.content_type, status=403
                )


def pdp_protect(view):
    """ Wrapper function for views that shall be protected by PDP authorization.
        This function can be used as a decorator of a view function, or as a
        modifier to be used in the url configuration file.
        e.g:
        ::

            urlpatterns = patterns('',
                ...
                url(r'^ows', pdp_protect(ows)),
                ...
            )
    """

    @functools.wraps(view)
    def wrapped(request, *args, **kwargs):
        pdp = getPDP()
        if pdp:
            try:
                authorized = pdp.authorize(request)
                message = "Not authorized"
                code = "NotAuthorized"
            except AuthorisationException as e:
                authorized = False
                message = str(e)
                code = e.code

            if not authorized:
                encoder = OWS20ExceptionXMLEncoder()
                return HttpResponse(
                    encoder.serialize(
                        encoder.encode_exception(message, "2.0.0", code)
                    ),
                    encoder.content_type, status=403
                )

        return view(request, *args, **kwargs)

    return wrapped



#-------------------------------------------------------------------------------
#
# simple tool comparing XML documents - powered by Python miniDOM
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
#

""" Simple XML documets' comparator. """

import xml.dom.minidom as dom

from django.utils.six import string_types


# define node types 
ELEMENT_NODE                = dom.Element.ELEMENT_NODE
ATTRIBUTE_NODE              = dom.Element.ATTRIBUTE_NODE
TEXT_NODE                   = dom.Element.TEXT_NODE
CDATA_SECTION_NODE          = dom.Element.CDATA_SECTION_NODE
ENTITY_REFERENCE_NODE       = dom.Element.ENTITY_REFERENCE_NODE
ENTITY_NODE                 = dom.Element.ENTITY_NODE
PROCESSING_INSTRUCTION_NODE = dom.Element.PROCESSING_INSTRUCTION_NODE
COMMENT_NODE                = dom.Element.COMMENT_NODE
DOCUMENT_NODE               = dom.Element.DOCUMENT_NODE
DOCUMENT_TYPE_NODE          = dom.Element.DOCUMENT_TYPE_NODE
DOCUMENT_FRAGMENT_NODE      = dom.Element.DOCUMENT_FRAGMENT_NODE
NOTATION_NODE               = dom.Element.NOTATION_NODE

# define note type to string conversion 
NODE_DICT = { 
    ELEMENT_NODE                : "ELEMENT_NODE",
    ATTRIBUTE_NODE              : "ATTRIBUTE_NODE",
    TEXT_NODE                   : "TEXT_NODE",
    CDATA_SECTION_NODE          : "CDATA_SECTION_NODE",
    ENTITY_REFERENCE_NODE       : "ENTITY_REFERENCE_NODE",
    ENTITY_NODE                 : "ENTITY_NODE",
    PROCESSING_INSTRUCTION_NODE : "PROCESSING_INSTRUCTION_NODE",
    COMMENT_NODE                : "COMMENT_NODE",
    DOCUMENT_NODE               : "DOCUMENT_NODE",
    DOCUMENT_TYPE_NODE          : "DOCUMENT_TYPE_NODE",
    DOCUMENT_FRAGMENT_NODE      : "DOCUMENT_FRAGMENT_NODE",
    NOTATION_NODE               : "NOTATION_NODE",
} 

# exceptions 

class XMLError( Exception ) : 
    """ XML base error error """

class XMLParseError( XMLError ) : 
    """ XML parse error """

class XMLMismatchError( XMLError ) : 
    """ XML mismatch error """ 

#-------------------------------------------------------------------------------
# low level  utilities 

def _getNodeName( node ) : 
    """ get full node name in '{namespace}tagName' format """
    if ( node.namespaceURI is None ) : 
        return node.nodeName
    else : 
        return  "{%s}%s"%( node.namespaceURI , node.localName ) 

def _packName( pair ) :
    """ pack the (<namespace>,<localname>) tuple to curly bracket notation 
        {<namespace>}<localname> """ 
    if ( pair[0] is None ) : 
        return pair[1] 
    else :
        return "{%s}%s"%(pair[0],pair[1])


def _skipIgnorable( node , path ) : 
    """ get node sibling skipping empty text nodes and comments """ 

    while ( node is not None ) :  
        # expected nodes - return immediatelly 
        if node.nodeType in (ELEMENT_NODE,CDATA_SECTION_NODE): break 
        # special treatment of text nodes - ignore blank text 
        if node.nodeType == TEXT_NODE : 
            # ignore blank text  
            if 0 < len( node.wholeText.strip() ) : break 
        # unexpected nodes - raise exception  
        if node.nodeType in (ATTRIBUTE_NODE,DOCUMENT_NODE,DOCUMENT_FRAGMENT_NODE,
                NOTATION_NODE,ENTITY_REFERENCE_NODE,ENTITY_NODE,DOCUMENT_TYPE_NODE): 
            raise XMLParseError("Unexpected child node '%s' ! PATH='%s'" % (NODE_DICT[node.nodeType],path)) 
        # the rest is just ignored  
        #if node.nodeType in (COMMENT_NODE,PROCESSING_INSTRUCTION_NODE) : pass
        node = node.nextSibling 

    return node 

def _compareAttributes( a0 , a1 , level , path , verbose = False ) : 

    # both nodes have no attributes 
    if ( a0 is None ) and ( a1 is None ) : return 

    #attribute mismatch  
    if ( a0 is None ) or ( a1 is None ) :
            raise XMLMismatchError("Attribute mismatch! PATH=\"%s\""%path) 

    # get list of attributes and filter-out namespace definitions 
    isNotNS  = lambda v : ( v[0][0] != "http://www.w3.org/2000/xmlns/" )  
    packName = lambda v : ( _packName(v[0]) , v[1].strip() )

    items0 = sorted( map( packName , filter( isNotNS , a0.itemsNS() ) ) )  
    items1 = sorted( map( packName , filter( isNotNS , a1.itemsNS() ) ) )  

    if len( items0 ) != len( items0 ) : 
        if verbose : 
            for item in items0 :
                print (" < \t %s@%s=\"%s\"" %( path , item[0] , item[1] ))
            for item in items1 :
                print (" > \t %s@%s=\"%s\"" %( path , item[0] , item[1] ))
        raise XMLMismatchError("Attribute count mismatch! PATH=\"%s\""%path)

    for pair in zip( items0 , items1 ) : 
        if verbose : 
            print (" < \t %s@%s=\"%s\"" %( path , pair[0][0] , pair[0][1] )) 
            print (" > \t %s@%s=\"%s\"" %( path , pair[1][0] , pair[1][1] ))
        if ( pair[0] != pair[1]) : 
            raise XMLMismatchError("Attribute mismatch! PATH=\"%s\""%path)


def _compareNode( n0 , n1 , level = 0 , path = "/" , verbose = False ) : 
    """ compare DOM node or element subtree """
    
    #nn0 , nn1 = _getNodeName( n0 ), _getNodeName( n1 )
    nn0 , nn1 = n0.nodeName, n1.nodeName

    path0 = "%s/%s"%( path , nn0 ) if level > 1 else "/%s"%nn0 if level == 1 else _getNodeName( n0 )  
    path1 = "%s/%s"%( path , nn1 ) if level > 1 else "/%s"%nn1 if level == 1 else _getNodeName( n0 )

    if verbose : 
        print ("< \t %s" %( path0 )) 
        print ("> \t %s" %( path1 ))
    
    # compare node name and node type 
    if (( n0.nodeType != n1.nodeType ) 
        or ( _getNodeName( n0 ) != _getNodeName( n1 ) )): 
            raise XMLMismatchError("Node mismatch! PATH0=\"%s\" vs. PATH1=\"%s\""%(path0,path1)) 

    # compare attributes 
    _compareAttributes( n0.attributes , n1.attributes , level , path0 , verbose ) 

    # in case of text-nodes and CDATA section check the content 
    if n0.nodeType == TEXT_NODE : 
        if verbose : 
            print (" < TEXT: \t \"%s\"" % n0.wholeText.strip())
            print (" > TEXT: \t \"%s\"" % n1.wholeText.strip())
        if n0.wholeText.strip() != n1.wholeText.strip() : 
            raise XMLMismatchError("Text mismatch! PATH=\"%s\""%(path)) 
        return 

    if n0.nodeType == CDATA_SECTION_NODE : 
        if verbose : 
            print (" < CDATA: \t \"%s\"" % n0.wholeText)
            print (" > CDATA: \t \"%s\"" % n1.wholeText)
        if n0.wholeText != n1.wholeText : 
            raise XMLMismatchError("CDATA mismatch! PATH=\"%s\""%(path)) 
        return 

    
    # get first child
    nn0 = _skipIgnorable( n1.firstChild , path )
    nn1 = _skipIgnorable( n0.firstChild , path )
    while ( nn0 is not None ) and ( nn1 is not None ) : 
        # sublevel comparison 
        _compareNode( nn0 , nn1 , level+1 , path0 , verbose ) 
        #get next sibling 
        nn0 = _skipIgnorable( nn0.nextSibling , path )
        nn1 = _skipIgnorable( nn1.nextSibling , path )

    # make sure there are no remaining nodes 
    if not (( nn0 is None ) and ( nn1 is None )) :  
        raise XMLMismatchError("Childern count mismatch! PATH=\"%s\""%path0)

#-------------------------------------------------------------------------------

def xmlCompareDOMs( xml0 , xml1  , verbose = False ) : 
    """ Compare two XML documents passed as DOM trees (xml.dom.minidom)."""

    return _compareNode( xml0 , xml1 , verbose = verbose ) 


def xmlCompareStrings( str0 , str1  , verbose = False ) : 
    """ Compare two XML documents passed as strings. """ 

    def parse( src , label ) : 
        try : 
            return dom.parseString( src ) 
        except Exception as e : 
            raise XMLParseError("Failed to parse %s XML string! %s" % ( label , str(e) )) 

    return xmlCompareDOMs( parse(str0,"the first") , parse(str1,"the second") , verbose ) 


def xmlCompareFiles( src0 , src1 , verbose = False ) : 
    """ Compare two XML documents passed as filenames, file or file-like objects.""" 

    def parseFileName( src ) : 
        try : 
            with open( src ) as fid : 
                return dom.parse( fid ) 
        except Exception as e : 
            raise XMLParseError("Failed to parse the \"%s\" file! %s" % ( src , str(e) )) 

    def parseFileObj( src , label ) : 
        try : 
            return dom.parse( src ) 
        except Exception as e : 
            raise XMLParseError("Failed to parse the %s XML file(-like) object! %e" % ( label , str(e) )) 

    def parse( src , label ) : 
        return  parseFileName( src ) if ( type(src) in string_types ) else parseFileObj( src , label ) 

    return xmlCompareDOMs( parse(src0,"the first") , parse(src1,"the second") , verbose ) 

#-------------------------------------------------------------------------------




# -----------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# -----------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

from django.db.models import Q

from eoxserver.core.decoders import xml, kvp, typelist, lower
from eoxserver.resources.coverages import models
from eoxserver.render.coverage.objects import from_model
from eoxserver.services.ows.wcs.basehandlers import (
    WCSGetCapabilitiesHandlerBase
)
from eoxserver.services.ows.wcs.v20.util import nsmap, SectionsMixIn
from eoxserver.services.ows.wcs.v20.parameters import (
    WCS20CapabilitiesRenderParams
)
from eoxserver.services.ecql import parse, to_filter
from eoxserver.services import filters


class WCS20GetCapabilitiesHandler(WCSGetCapabilitiesHandlerBase):
    versions = ("2.0.0", "2.0.1")
    methods = ['GET', 'POST']

    additional_parameters = {
        'cql': None
    }

    def get_decoder(self, request):
        if request.method == "GET":
            return WCS20GetCapabilitiesKVPDecoder(request.GET)
        elif request.method == "POST":
            return WCS20GetCapabilitiesXMLDecoder(request.body)

    def lookup_coverages(self, decoder):
        sections = decoder.sections
        inc_coverages = (
            "all" in sections or "contents" in sections or
            "coveragesummary" in sections
        )
        inc_dataset_series = (
            "all" in sections or "contents" in sections or
            "datasetseriessummary" in sections
        )

        if inc_coverages:
            cql_text = decoder.cql
            if cql_text:
                qs = models.EOObject.objects.all()
                mapping, mapping_choices = filters.get_field_mapping_for_model(
                    models.Coverage  # TODO: allow mapping to Mosaic as-well?
                )
                ast = parse(cql_text)
                filter_expressions = to_filter(ast, mapping, mapping_choices)
                qs = qs.filter(filter_expressions)

            else:
                qs = models.EOObject.objects.filter(
                    service_visibility__service='wcs',
                    service_visibility__visibility=True
                )

            qs = qs.order_by('id').select_subclasses(
                models.Coverage, models.Mosaic
            )

            coverages = [
                from_model(coverage)
                for coverage in qs
            ]
        else:
            coverages = []

        if inc_dataset_series:
            cql_text = decoder.datasetseriescql
            if cql_text:
                qs = models.EOObject.objects.all()
                mapping, mapping_choices = filters.get_field_mapping_for_model(
                    qs.model  # TODO: mapping to Collection/Product would be better
                )
                ast = parse(cql_text)
                filter_expressions = to_filter(ast, mapping, mapping_choices)
                qs = qs.filter(filter_expressions)

            dataset_series_qs = models.EOObject.objects.filter(
                Q(
                    product__isnull=False,
                    service_visibility__service='wcs',
                    service_visibility__visibility=True
                ) | Q(
                    collection__isnull=False
                )
            ).exclude(
                collection__isnull=False,
                service_visibility__service='wcs',
                service_visibility__visibility=False
            )

            # reduce data transfer by only selecting required elements
            dataset_series_qs = dataset_series_qs.only(
                "identifier", "begin_time", "end_time", "footprint"
            )

        else:
            dataset_series_qs = models.EOObject.objects.none()

        return coverages, dataset_series_qs

    def get_params(self, models, decoder):
        coverages, dataset_series = models
        return WCS20CapabilitiesRenderParams(
            coverages, dataset_series, decoder.sections,
            decoder.acceptlanguages, decoder.acceptformats,
            decoder.updatesequence
        )


class WCS20GetCapabilitiesKVPDecoder(kvp.Decoder, SectionsMixIn):
    sections            = kvp.Parameter(type=typelist(lower, ","), num="?", default=["all"])
    updatesequence      = kvp.Parameter(num="?")
    acceptversions      = kvp.Parameter(type=typelist(str, ","), num="?")
    acceptformats       = kvp.Parameter(type=typelist(str, ","), num="?", default=["text/xml"])
    acceptlanguages     = kvp.Parameter(type=typelist(str, ","), num="?")
    cql                 = kvp.Parameter(num="?")
    datasetseriescql    = kvp.Parameter(num="?")


class WCS20GetCapabilitiesXMLDecoder(xml.Decoder, SectionsMixIn):
    sections            = xml.Parameter("ows:Sections/ows:Section/text()", num="*", default=["all"])
    updatesequence      = xml.Parameter("@updateSequence", num="?")
    acceptversions      = xml.Parameter("ows:AcceptVersions/ows:Version/text()", num="*")
    acceptformats       = xml.Parameter("ows:AcceptFormats/ows:OutputFormat/text()", num="*", default=["text/xml"])
    acceptlanguages     = xml.Parameter("ows:AcceptLanguages/ows:Language/text()", num="*")

    # TODO: find suitable place in XML to pass CQL queries
    cql                 = None
    datasetseriescql    = None

    namespaces = nsmap



# By Christian Muirhead, Menno Smits and Michael Foord 2008
# WTF license
# http://voidspace.org.uk/blog

"""
``total_ordering`` and ``force_total_ordering`` are class decorators for
Python 2.6 & Python 3.

They provides *all* the rich comparison methods on a class by defining *any*
one of '__lt__', '__gt__', '__le__', '__ge__'.

``total_ordering`` fills in all unimplemented rich comparison methods, assuming
at least one is implemented. ``__lt__`` is taken as the base comparison method
on which the others are built, but if that is not available it will be
constructed from the first one found.

``force_total_ordering`` does the same, but having taken a comparison method as
the base it fills in *all* the others - this overwrites additional comparison
methods that may be implemented, guaranteeing consistent comparison semantics.

::

    from total_ordering import total_ordering

    @total_ordering
    class Something(object):
        def __init__(self, value):
            self.value = value
        def __lt__(self, other):
            return self.value < other.value

It also works with Python 2.5, but you need to do the wrapping yourself:

::

    from total_ordering import total_ordering

    class Something(object):
        def __init__(self, value):
            self.value = value
        def __lt__(self, other):
            return self.value < other.value

    total_ordering(Something)

It would be easy to modify for it to work as a class decorator for Python
3.X and a metaclass for Python 2.X.
"""

import sys as _sys

if _sys.version_info[0] == 3:
    def _has_method(cls, name):
        for B in cls.__mro__:
            if B is object:
                continue
            if name in B.__dict__:
                return True
        return False
else:
    def _has_method(cls, name):
        for B in cls.mro():
            if B is object:
                continue
            if name in B.__dict__:
                return True
        return False


def _ordering(cls, overwrite):
    def setter(name, value):
        if overwrite or not _has_method(cls, name):
            value.__name__ = name
            setattr(cls, name, value)

    comparison = None
    if not _has_method(cls, '__lt__'):
        for name in 'gt le ge'.split():
            if not _has_method(cls, '__' + name + '__'):
                continue
            comparison = getattr(cls, '__' + name + '__')
            if name.endswith('e'):
                eq = lambda s, o: comparison(s, o) and comparison(o, s)
            else:
                eq = lambda s, o: not comparison(s, o) and not comparison(o, s)
            ne = lambda s, o: not eq(s, o)
            if name.startswith('l'):
                setter('__lt__', lambda s, o: comparison(s, o) and ne(s, o))
            else:
                setter('__lt__', lambda s, o: comparison(o, s) and ne(s, o))
            break
        assert comparison is not None, (
            'must have at least one of ge, gt, le, lt'
        )

    setter('__ne__', lambda s, o: s < o or o < s)
    setter('__eq__', lambda s, o: not s != o)
    setter('__gt__', lambda s, o: o < s)
    setter('__ge__', lambda s, o: not (s < o))
    setter('__le__', lambda s, o: not (s > o))
    return cls


def total_ordering(cls):
    return _ordering(cls, False)


def force_total_ordering(cls):
    return _ordering(cls, True)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-02-06 15:13
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('services', '0002_service_visibility_cardinality_uniqueness'),
    ]

    operations = [
        migrations.AlterField(
            model_name='servicevisibility',
            name='service',
            field=models.CharField(choices=[('wms', 'WMS'), ('wcs', 'WCS'), ('os', 'OpenSearch'), ('wc', 'WebClient')], max_length=4),
        ),
    ]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import sys

from django.utils.six import reraise


ZERO_OR_ONE = "?"
ONE_OR_MORE = "+"
ANY = "*"

SINGLE_VALUES = (ZERO_OR_ONE, 1)


class DecodingException(Exception):
    """ Base Exception class to be thrown whenever a decoding failed.
    """

    def __init__(self, message, locator=None):
        super(DecodingException, self).__init__(message)
        self.locator = locator


class WrongMultiplicityException(DecodingException):
    """ Decoding Exception to be thrown when the multiplicity of a parameter did
        not match the expected count.
    """

    code = "InvalidParameterValue"

    def __init__(self, locator, expected, result):
        super(WrongMultiplicityException, self).__init__(
            "Parameter '%s': expected %s got %d" % (locator, expected, result),
            locator
        )


class MissingParameterException(DecodingException):
    """ Exception to be thrown, when a decoder could not read one parameter,
        where exactly one was required.
    """
    code = "MissingParameterValue"

    def __init__(self, locator):
        super(MissingParameterException, self).__init__(
            "Missing required parameter '%s'" % locator, locator
        )


class MissingParameterMultipleException(DecodingException):
    """ Exception to be thrown, when a decoder could not read at least one
        parameter, where one ore more were required.
    """
    code = "MissingParameterValue"

    def __init__(self, locator):
        super(MissingParameterMultipleException, self).__init__(
            "Missing at least one required parameter '%s'" % locator, locator
        )


class NoChoiceResultException(DecodingException):
    pass


class ExclusiveException(DecodingException):
    pass

# NOTE: The following exceptions may get propagated as OWS exceptions
#       therefore it is necessary to set the proper OWS exception code.


class InvalidParameterException(DecodingException):
    code = "InvalidParameterValue"


# Compound fields


class Choice(object):
    """ Tries all given choices until one does return something.
    """

    def __init__(self, *choices):
        self.choices = choices

    def __get__(self, decoder, decoder_class=None):
        for choice in self.choices:
            try:
                return choice.__get__(decoder, decoder_class)
            except:
                continue
        raise NoChoiceResultException


class Exclusive(object):
    """ For mutual exclusive Parameters.
    """

    def __init__(self, *choices):
        self.choices = choices

    def __get__(self, decoder, decoder_class=None):
        result = None
        num = 0
        for choice in self.choices:
            try:
                result = choice.__get__(decoder, decoder_class)
                num += 1
            except:
                continue

        if num != 1:
            raise ExclusiveException

        return result


class Concatenate(object):
    """ Helper to concatenate the results of all sub-parameters to one.
    """
    def __init__(self, *choices, **kwargs):
        self.choices = choices
        self.allow_errors = kwargs.get("allow_errors", True)

    def __get__(self, decoder, decoder_class=None):
        result = []
        for choice in self.choices:
            try:
                value = choice.__get__(decoder, decoder_class)
                if isinstance(value, (list, tuple)):
                    result.extend(value)
                else:
                    result.append(value)
            except Exception:
                if self.allow_errors:
                    # swallow exception
                    continue

                exc_info = sys.exc_info()
                reraise(exc_info[0], exc_info[1], exc_info[2])

        return result


# Type conversion helpers


class typelist(object):
    """ Helper for XMLDecoder schemas that expect a string that represents a
        list of a type separated by some separator.
    """

    def __init__(self, typ=None, separator=" "):
        self.typ = typ
        self.separator = separator

    def __call__(self, value):
        split = value.split(self.separator)
        if self.typ:
            return [self.typ(v) for v in split]
        return split


class fixed(object):
    """ Helper for parameters that are expected to be have a fixed value and
        raises a ValueError if not.
    """

    def __init__(self, value, case_sensitive=True):
        self.value = value if case_sensitive else value.lower()
        self.case_sensitive = case_sensitive

    def __call__(self, value):
        compare = value if self.case_sensitive else value.lower()
        if self.value != compare:
            raise ValueError("Value mismatch, expected %s, got %s." %
                (self.value, value)
            )

        return value


class enum(object):
    """ Helper for parameters that are expected to be in a certain enumeration.
        A ValueError is raised if not.
    """

    def __init__(self, values, case_sensitive=True, error_class=ValueError):
        self.values = values
        self.compare_values = values if case_sensitive else [lower(v) for v in values]
        self.case_sensitive = case_sensitive
        self.error_class = error_class

    def __call__(self, value):
        compare = value if self.case_sensitive else value.lower()
        if compare not in self.compare_values:
            raise self.error_class("Unexpected value '%s'. Expected one of: %s." %
                (value, ", ".join(map(lambda s: "'%s'" % s, self.values)))
            )

        return value


def lower(value):
    """ Functor to return a lower-case string.
    """
    return value.lower()


def upper(value):
    """ Functor to return a upper-case string.
    """
    return value.upper()


def strip(value):
    """ Functor to return a whitespace stripped string.
    """
    return value.strip()


class value_range(object):
    """ Helper to assert that a given parameter is within a specified range.
    """

    def __init__(self, min, max, type=float):
        self._min = min
        self._max = max
        self._type = type

    def __call__(self, raw):
        value = self._type(raw)
        if value < self._min or value > self._max:
            raise ValueError(
                "Given value '%s' exceeds expected bounds (%s, %s)"
                % (value, self._min, self._max)
            )
        return value


def boolean(raw):
    """ Functor to convert "true"/"false" to a boolean.
    """
    raw = raw.lower()
    if not raw in ("true", "false"):
        raise ValueError("Could not parse a boolean value from '%s'." % raw)
    return raw == "true"


def to_dict(decoder, dict_class=dict):
    """ Utility function to get a dictionary representation of the given decoder.
        This function invokes all decoder parameters and sets the dictionary
        fields accordingly
    """
    return dict(
        (name, getattr(decoder, name))
        for name in dir(decoder)
        if not name.startswith("_") and name != "namespaces"
    )



# -----------------------------------------------------------------------------
#
# Project: ngEO Browse Server <http://ngeo.eox.at>
# Authors: Stephan Meissl <stephan.meissl@eox.at>
#
# -----------------------------------------------------------------------------
# Copyright (C) 2018 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import logging
from os import remove
from os.path import abspath, exists
import traceback
from sqlite3 import dbapi2
import textwrap

from django.core.management.base import BaseCommand, CommandError
from django.utils.dateparse import parse_datetime, parse_time
from django.contrib.gis.db.models import Extent

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


logger = logging.getLogger(__name__)


def decoder(conv_func):
    """
    Convert bytestrings from Python's sqlite3 interface to a regular string.
    """
    return lambda s: conv_func(s.decode())


dbapi2.register_converter("bool", b'1'.__eq__)
dbapi2.register_converter("time", decoder(parse_time))
dbapi2.register_converter("datetime", decoder(parse_datetime))
dbapi2.register_converter("timestamp", decoder(parse_datetime))
dbapi2.register_converter("TIMESTAMP", decoder(parse_datetime))


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    help = ("Synchronizes the MapCache SQLite DB holding times and extents.")

    def add_arguments(self, parser):
        sync_parser = self.add_subparser(parser, 'sync')
        sync_parser.add_argument(
            "--force", "-f",
            dest="force", action="store_true", default=False,
            help=(
                "Optional. Force the re-generation of the database file if "
                "it already exists."
            )
        )
        sync_parser.add_argument(
            "--unique-times", "-u",
            dest="unique_times", action="store_true", default=False,
            help=(
                "Optional. Force unique time entries. Extents are merged."
            )
        )
        sync_parser.add_argument(
            "--no-index",
            dest="index", action="store_false", default=True,
            help=(
                "Optional. Do not create an index."
            )
        )

    def handle(self, subcommand, *args, **kwargs):
        """ Dispatch sub-commands: register, deregister.
        """
        if subcommand == "sync":
            self.handle_sync(*args, **kwargs)

    def handle_sync(self, force, unique_times=False, index=True, **kwargs):
        # parse command arguments
        self.verbosity = int(kwargs.get("verbosity", 1))

        logger.info(
            "Starting synchronization of MapCache SQLite DB holding "
            "times and extents."
        )

        create_sql = textwrap.dedent('''
            CREATE TABLE "time" (
                "start_time" timestamp with time zone NOT NULL,
                "end_time" timestamp with time zone NOT NULL,
                "minx" double precision NOT NULL,
                "miny" double precision NOT NULL,
                "maxx" double precision NOT NULL,
                "maxy" double precision NOT NULL
            )
        ''')

        for collection in models.Collection.objects.all():
            db_path = abspath("%s.sqlite" % collection.identifier)
            if exists(db_path):
                if force:
                    logger.info("Deleting old MapCache SQLite file.")
                    remove(db_path)
                else:
                    logger.error("MapCache SQLite exists.")
                    raise CommandError("MapCache SQLite exists.")

            logger.info("Creating new MapCache SQLite file.")
            conn = dbapi2.connect(
                db_path,
                detect_types=dbapi2.PARSE_DECLTYPES | dbapi2.PARSE_COLNAMES,
            )
            conn.execute(create_sql)

            self.handle_collection(conn, collection, unique_times)

            if index:
                conn.execute(
                    'CREATE INDEX time_idx ON time '
                    '(start_time, end_time, minx, miny, maxx, maxy)'
                )

            conn.commit()
            conn.close()

        logger.info(
            "Finished generation of MapCache SQLite DB holding times "
            "and extents."
        )

    def handle_collection(self, conn, collection, unique_times):
        try:
            logger.info("Syncing layer '%s'" % collection.identifier)

            logger.debug("Starting query for browses")

            products_qs = models.Product.objects.filter(
                collections=collection,
            ).annotate(
                extent=Extent('footprint')
            ).values(
                'begin_time', 'end_time', 'extent'
            ).order_by(
                'begin_time', 'end_time'
            )

            logger.info("Number products: %s" % products_qs.count())

            if not unique_times:
                time_intervals = (
                    (
                        product['begin_time'],
                        product['end_time']
                    ) + product['extent']
                    for product in products_qs
                )
            else:
                logger.debug("Starting query for unique times")
                # optimization for when there are a lot of equal time entries
                # like for Sentinel-2
                unique_times_qs = models.Product.objects.filter(
                    collections=collection,
                ).values_list(
                    'begin_time', 'end_time'
                ).distinct(
                    'begin_time', 'end_time'
                ).order_by(
                    'begin_time', 'end_time'
                )
                logger.info(
                    "Number unique times: %s" % unique_times_qs.count()
                )

                logger.info("Iterating through unique times")
                time_intervals = []
                i = 1
                for begin_time, end_time in unique_times_qs:
                    logger.debug(
                        "Working on unique time %s: %s/%s " %
                        (i, begin_time, end_time)
                    )
                    i += 1

                    minx, miny, maxx, maxy = (None,) * 4

                    # search for all browses within that time interval and
                    # combine extent
                    time_qs = products_qs.filter(
                        begin_time=begin_time,
                        end_time=end_time
                    )

                    if time_qs.count() <= 0:
                        logger.errro(
                            "DB queries got different results which should "
                            "never happen."
                        )
                        raise CommandError("DB queries got different results.")
                    else:
                        for time in time_qs:
                            # decode extent from the above hack
                            minx_tmp, miny_tmp, maxx_tmp, maxy_tmp = \
                                time['extent']
                            # change one extent to ]0,360] if difference gets
                            # smaller
                            if minx is not None and maxx is not None:
                                if (minx_tmp <= 0 and maxx_tmp <= 0 and
                                        (minx-maxx_tmp) > (360+minx_tmp-maxx)):
                                    minx_tmp += 360
                                    maxx_tmp += 360
                                elif (minx <= 0 and maxx <= 0 and
                                        (minx_tmp-maxx) > (360+minx-maxx_tmp)):
                                    minx += 360
                                    maxx += 360
                            minx = min(
                                i for i in [minx_tmp, minx] if i is not None
                            )
                            miny = min(
                                i for i in [miny_tmp, miny] if i is not None
                            )
                            maxx = max(
                                i for i in [maxx_tmp, maxx] if i is not None
                            )
                            maxy = max(
                                i for i in [maxy_tmp, maxy] if i is not None
                            )

                    # check if previous element in ordered list overlaps
                    if (
                        len(time_intervals) > 0 and (
                            (
                                (
                                    begin_time == end_time or
                                    time_intervals[-1][0] ==
                                    time_intervals[-1][1]
                                ) and (
                                    time_intervals[-1][0] <= end_time and
                                    time_intervals[-1][1] >= begin_time
                                )
                            ) or (
                                time_intervals[-1][0] < end_time and
                                time_intervals[-1][1] > begin_time
                            )
                        )
                    ):
                        begin_time = min(begin_time, time_intervals[-1][0])
                        end_time = max(end_time, time_intervals[-1][1])
                        minx = min(minx, time_intervals[-1][2])
                        miny = min(miny, time_intervals[-1][3])
                        maxx = max(maxx, time_intervals[-1][4])
                        maxy = max(maxy, time_intervals[-1][5])
                        time_intervals.pop(-1)
                    time_intervals.append(
                        (begin_time, end_time, minx, miny, maxx, maxy)
                    )

                logger.info(
                    "Number non-overlapping time intervals: %s" %
                    len(time_intervals)
                )

            logger.info(
                "Starting saving time intervals to MapCache SQLite file"
            )

            conn.executemany(
                'INSERT INTO time VALUES (?, ?, ?, ?, ?, ?)',
                time_intervals
            )

            logger.info(
                "Finished saving time intervals to MapCache SQLite file"
            )

        except Exception as e:
            logger.error(
                "Failure during generation of MapCache SQLite DB."
            )
            logger.error(
                "Exception was '%s': %s" % (type(e).__name__, str(e))
            )
            logger.debug(traceback.format_exc() + "\n")



from .getcapabilities import WCS11GetCapabilitiesHandler
from .describecoverage import WCS11DescribeCoverageHandler
from .getcoverage import WCS11GetCoverageHandler


GetCapabilitiesHandler = WCS11GetCapabilitiesHandler
DescribeCoverageHandler = WCS11DescribeCoverageHandler
GetCoverageHandler = WCS11GetCoverageHandler



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from eoxserver.core.decoders import InvalidParameterException
from eoxserver.core.util.timetools import parse_iso8601


def parse_bbox(string):
    try:
        bbox = [float(v) for v in string.split(",")]
    except ValueError:
        raise InvalidParameterException("Invalid 'BBOX' parameter.", "bbox")


    if len(bbox) != 4:
        raise InvalidParameterException(
            "Wrong number of arguments for 'BBOX' parameter.", "bbox"
        )

    return bbox


def parse_time(string):
    items = string.split("/")

    if len(items) == 1:
        return [parse_iso8601(items[0])]
    elif len(items) in (2, 3):
        # ignore resolution
        return [parse_iso8601(items[0]), parse_iso8601(items[1])]

    raise InvalidParameterException("Invalid TIME parameter.", "time")


def int_or_str(string):
    try:
        return int(string)
    except ValueError:
        return string



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class HTTPMethodNotAllowedError(Exception):
    """ This exception is raised in case of a HTTP requires with unsupported
    HTTP method.
    This exception should always lead to the 405 Method not allowed HTTP error.

    The constructor takes two arguments, the error message ``mgs`` and the list
    of the accepted HTTP methods ``allowed_methods``.
    """
    def __init__(self, msg, allowed_methods):
        super(HTTPMethodNotAllowedError, self).__init__(msg)
        self.allowed_methods = allowed_methods


class InvalidRequestException(Exception):
    """
    This exception indicates that the request was invalid and an exception
    report shall be returned to the client.

    The constructor takes three arguments, namely ``msg``, the error message,
    ``code``, the error code, and ``locator``, which is needed in OWS
    exception reports for indicating which part of the request produced the
    error.

    How exactly the exception reports are constructed is not defined by the
    exception, but by exception handlers.
    """
    def __init__(self, msg, code=None, locator=None):
        super(InvalidRequestException, self).__init__(msg)

        self.code = code or "InvalidRequest"
        self.locator = locator

    def __str__(self):
        return "Invalid Request: Code: %s; Locator: %s; Message: '%s'" % (
            self.code, self.locator,
            super(InvalidRequestException, self).__str__()
        )


class VersionNegotiationException(Exception):
    """
    This exception indicates that version negotiation fails. Such errors can
    happen with OWS 2.0 compliant "new-style" version negotation.
    """
    code = "VersionNegotiationFailed"

    def __str__(self):
        return "Version negotiation failed."


class LocatorListException(Exception):
    """ Base class for exceptions that report that a number of items are
        missing or invalid
    """
    def __init__(self, items):
        self.items = items

    @property
    def locator(self):
        "This property provides a list of all missing/invalid items."
        return " ".join(self.items)


class InvalidAxisLabelException(Exception):
    """
    This exception indicates that an invalid axis name was chosen in a WCS
    2.0 subsetting parameter.
    """
    code = "InvalidAxisLabel"

    def __init__(self, axis_label):
        super(InvalidAxisLabelException, self).__init__(
            "Invalid axis label: '%s'." % axis_label
        )
        self.locator = axis_label


class InvalidSubsettingException(Exception):
    """
    This exception indicates an invalid WCS 2.0 subsetting parameter was
    submitted.
    """
    code = "InvalidSubsetting"
    locator = "subset"


class InvalidSubsettingCrsException(Exception):
    """
    This exception indicates an invalid WCS 2.0 subsettingCrs parameter was
    submitted.
    """
    code = "SubsettingCrs-NotSupported"
    locator = "subsettingCrs"


class InvalidOutputCrsException(Exception):
    """
    This exception indicates an invalid WCS 2.0 outputCrs parameter was
    submitted.
    """
    code = "OutputCrs-NotSupported"
    locator = "outputCrs"


class NoSuchCoverageException(LocatorListException):
    """ This exception indicates that the requested coverage(s) do not
        exist.
    """
    code = "NoSuchCoverage"

    def __str__(self):
        return "No such Coverage%s with ID: %s" % (
            "" if len(self.items) == 1 else "s",
            ", ".join(map(lambda i: "'%s'" % i, self.items))
        )


class NoSuchDatasetSeriesOrCoverageException(LocatorListException):
    """ This exception indicates that the requested coverage(s) or dataset
        series do not exist.
    """
    code = "NoSuchDatasetSeriesOrCoverage"

    def __str__(self):
        return "No such Coverage%s or Dataset Series with ID: %s" % (
            " " if len(self.items) == 1 else "s",
            ", ".join(map(lambda i: "'%s'" % i, self.items))
        )


class OperationNotSupportedException(Exception):
    """ Exception to be thrown when some operations are not supported or
        disabled.
    """
    def __init__(self, message, operation=None):
        super(OperationNotSupportedException, self).__init__(message)
        self.operation = operation

    @property
    def locator(self):
        return self.operation

    code = "OperationNotSupported"


class ServiceNotSupportedException(OperationNotSupportedException):
    """ Exception to be thrown when a specific OWS service is not enabled.
    """
    def __init__(self, service):
        self.service = service

    def __str__(self):
        if self.service:
            return "Service '%s' is not supported." % self.service.upper()
        else:
            return "Service is not supported."


class VersionNotSupportedException(Exception):
    """ Exception to be thrown when a specific OWS service version is not
        supported.
    """
    def __init__(self, service, version):
        self.service = service
        self.version = version

    def __str__(self):
        if self.service:
            return "Service '%s' version '%s' is not supported." % (
                self.service, self.version
            )
        else:
            return "Version '%s' is not supported." % self.version

    code = "InvalidParameterValue"


class InterpolationMethodNotSupportedException(Exception):
    """
    This exception indicates a not supported interpolation method.
    """
    code = "InterpolationMethodNotSupported"
    locator = "interpolation"


class RenderException(Exception):
    """ Rendering related exception.
    """
    def __init__(self, message, locator, is_parameter=True):
        super(RenderException, self).__init__(message)
        self.locator = locator
        self.is_parameter = is_parameter

    @property
    def code(self):
        return (
            "InvalidParameterValue" if self.is_parameter else "InvalidRequest"
        )


class NoSuchFieldException(Exception):
    """ Error in RangeSubsetting when band does not exist.
    """

    code = "NoSuchField"

    def __init__(self, msg, locator):
        super(NoSuchFieldException, self).__init__(msg)
        self.locator = locator


class InvalidFieldSequenceException(Exception):
    """ Error in RangeSubsetting for illegal intervals.
    """
    code = "InvalidFieldSequence"

    def __init__(self, msg, locator):
        super(NoSuchFieldException, self).__init__(msg)
        self.locator = locator


class InvalidScaleFactorException(Exception):
    """ Error in ScaleFactor and ScaleAxis operations
    """
    code = "InvalidScaleFactor"

    def __init__(self, scalefactor):
        super(InvalidScaleFactorException, self).__init__(
            "Scalefactor '%s' is not valid" % scalefactor
        )
        self.locator = scalefactor


class InvalidScaleExtentException(Exception):
    """ Error in ScaleExtent operations
    """
    code = "InvalidExtent"

    def __init__(self, low, high):
        super(InvalidScaleExtentException, self).__init__(
            "ScaleExtent '%s:%s' is not valid" % (low, high)
        )
        self.locator = high


class ScaleAxisUndefinedException(Exception):
    """ Error in all scaling operations involving an axis
    """

    code = "ScaleAxisUndefined"

    def __init__(self, axis):
        super(ScaleAxisUndefinedException, self).__init__(
            "Scale axis '%s' is undefined" % axis
        )
        self.locator = axis



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from eoxserver.core.decoders import kvp
from eoxserver.resources.coverages import crss
from eoxserver.services.ows.wms.util import parse_bbox
from eoxserver.services.ows.wms.exceptions import InvalidCRS
from eoxserver.services.ows.wms.basehandlers import (
    WMSBaseGetCapabilitiesHandler, WMSBaseGetMapHandler, WMSBaseGetMapDecoder
)
from eoxserver.services.ows.wms.v13.encoders import WMS13Encoder


class WMS13GetCapabilitiesHandler(WMSBaseGetCapabilitiesHandler):
    versions = ("1.3", "1.3.0")

    def get_encoder(self):
        return WMS13Encoder()


class WMS13GetMapHandler(WMSBaseGetMapHandler):
    service = ("WMS", None)
    versions = ("1.3.0", "1.3")

    def get_decoder(self, request):
        return WMS13GetMapDecoder(request.GET)


class WMS13GetMapDecoder(WMSBaseGetMapDecoder):
    _bbox = kvp.Parameter('bbox', type=parse_bbox, num=1)

    @property
    def bbox(self):
        bbox = self._bbox
        crs = self.crs
        srid = crss.parseEPSGCode(
            self.crs, (crss.fromShortCode, crss.fromURN, crss.fromURL)
        )
        if srid is None:
            raise InvalidCRS(crs, "crs")

        if crss.hasSwappedAxes(srid):
            miny, minx, maxy, maxx = bbox
        else:
            minx, miny, maxx, maxy = bbox

        return (minx, miny, maxx, maxy)

    crs = kvp.Parameter(num=1)

    srs = property(lambda self: self.crs)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2012 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from os.path import exists, join
from uuid import uuid4
import tempfile
import contextlib
import logging

import numpy

from eoxserver.contrib import gdal, gdal_array


logger = logging.getLogger(__name__)


def get_limits(dt):
    """ Returns the numeric limits of the GDAL/numpy datatype """
    if dt in gdal_array.codes:
        dt = gdal_array.codes[dt]

    try:
        info = numpy.iinfo(dt)
    except ValueError:
        info = numpy.finfo(dt)

    return info.min, info.max


#===============================================================================
# helper functions
#===============================================================================

def create_mem_copy(ds, *args, **kwargs):
    """ Create a new In-Memory Dataset as copy from an existing dataset. """
    mem_drv = gdal.GetDriverByName('MEM')
    return mem_drv.CreateCopy('', ds, *args, **kwargs)


def create_mem(sizex, sizey, numbands, datatype=gdal.GDT_Byte,
               options=None):
    """ Create a new In-Memory Dataset. """
    if options is None:
        options = []

    mem_drv = gdal.GetDriverByName('MEM')
    return mem_drv.Create('', sizex, sizey, numbands, datatype, options)


def create_temp(sizex, sizey, numbands, datatype=gdal.GDT_Byte,
                options=None, temp_root=None):
    """ Create a temporary Dataset. """

    temp_root = temp_root if temp_root is not None else tempfile.gettempdir()

    if options is None:
        options = []

    tiff_drv = gdal.GetDriverByName('GTiff')
    filename = join(temp_root, '%s.tif' % uuid4().hex)
    logger.debug(
        "Creating temporary dataset '%s' (%dx%dx%d)" % (
            filename, sizex, sizey, numbands
        )
    )
    return tiff_drv.Create(filename, sizex, sizey, numbands, datatype, options)


def cleanup_temp(ds):
    """ Delete a temporary dataset.
    """
    if isinstance(ds, str):
        driver = gdal.IdentifyDriver(ds)
        filelist = [ds]
    else:
        driver = ds.GetDriver()
        filelist = ds.GetFileList() or []
        ds.FlushCache()
        ds = None

    try:
        for filename in filelist:
            if filename.endswith(".ovr") or filename.endswith(".msk"):
                continue
            logger.debug("Cleaning up temporary dataset '%s'." % filename)
            driver.Delete(filename)
    except:
        logger.warning(
            "Failed to delete dataset '%s', falling back to raw delete."
            % filename
        )
        for filename in filelist:
            gdal.Unlink(filename)


@contextlib.contextmanager
def temporary_dataset(sizex, sizey, numbands, datatype=gdal.GDT_Byte,
                      options=None, temp_root=None):
    ds = create_temp(sizex, sizey, numbands, datatype, options, temp_root)
    yield ds
    cleanup_temp(ds)


def copy_projection(src_ds, dst_ds):
    """ Copy the projection and geotransform from on dataset to another """
    dst_ds.SetProjection(src_ds.GetProjection())
    dst_ds.SetGeoTransform(src_ds.GetGeoTransform())


def copy_metadata(src_ds, dst_ds):
    """ Copy the metadata from on dataset to another """
    dst_ds.SetMetadata(src_ds.GetMetadata_Dict())


def copy_nodatavalue(src_ds, dst_ds):
    """ Copy the nodata value from on dataset to another """
    assert src_ds.RasterCount == dst_ds.RasterCount
    for i in range(src_ds.RasterCount):
        src_band = src_ds.GetRasterBand(i+1)
        if src_band.GetNoDataValue() is not None:
            dst_band = dst_ds.GetRasterBand(i+1)
            dst_band.SetNoDataValue(src_band.GetNoDataValue())


def check_file_existence(filename):
    " Check if file exists and raise an IOError if it does. "
    if exists(filename):
        raise IOError("The output file '%s' already exists." % filename)



# -----------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Fabian Schindler <fabian.schindler@eox.at>
#
# -----------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import os
from setuptools import setup

# get version number
# from https://github.com/mapbox/rasterio/blob/master/setup.py#L55
init_py = os.path.join(os.path.dirname(__file__), 'eoxserver/__init__.py')
with open(init_py) as f:
    for line in f:
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            version = version.strip('"')
            version = version.strip("'")
            break


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def fullsplit(path, result=None):
    """
    Split a pathname into components (the opposite of os.path.join) in a
    platform-neutral way.
    """
    if result is None:
        result = []
    head, tail = os.path.split(path)
    if head == '':
        return [tail] + result
    if head == path:
        return result
    return fullsplit(head, [tail] + result)


packages, data_files = [], []
for dirpath, dirnames, filenames in os.walk('eoxserver'):
    for i, dirname in enumerate(dirnames):
        if dirname.startswith('.'):
            del dirnames[i]
    if '__init__.py' in filenames:
        packages.append('.'.join(fullsplit(dirpath)))
    elif filenames:
        data_files.append(
            [dirpath, [os.path.join(dirpath, f) for f in filenames]]
        )


# On readthecods.org we don't want the reftools to be build
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    ext_modules = []

setup(
    name='EOxServer',
    version=version.replace(' ', '-'),
    packages=packages,
    data_files=data_files,
    include_package_data=True,
    scripts=[
        "eoxserver/scripts/eoxserver-admin.py",
        "eoxserver/scripts/eoxserver-instance.py",
        "tools/eoxserver-atpd.py",
        "tools/eoxserver-validate_xml.py",
        "tools/eoxserver-preprocess.py"
    ],
    install_requires=[
        'django>=1.4',
        'python-dateutil',
        'django-model-utils<4.0.0',
        'zipstream',
        'psycopg2',
        'lxml',
        'pycql',
    ],
    extras_require={
        'dev': ['scipy'],
    },
    zip_safe=False,

    # Metadata
    author="EOX IT Services GmbH",
    author_email="office@eox.at",
    maintainer="EOX IT Services GmbH",
    maintainer_email="packages@eox.at",

    description="EOxServer is a server for Earth Observation (EO) data",
    long_description=read("README.rst"),

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Other Audience',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Database',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Visualization',
    ],

    license="EOxServer Open License (MIT-style)",
    keywords="Earth Observation, EO, OGC, WCS, WMS",
    url="http://eoxserver.org/"
)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from eoxserver.core import Component, implements
from eoxserver.services.ows.interfaces import ExceptionHandlerInterface
from eoxserver.services.ows.common.v20.encoders import OWS20ExceptionXMLEncoder
from eoxserver.core.decoders import (
    DecodingException, MissingParameterException
)


CODES_404 = frozenset((
    "NoSuchCoverage", "NoSuchDatasetSeriesOrCoverage", "InvalidAxisLabel",
    "InvalidSubsetting", "InterpolationMethodNotSupported", "NoSuchField",
    "InvalidFieldSequence", "InvalidScaleFactor", "InvalidExtent",
    "ScaleAxisUndefined", "SubsettingCrs-NotSupported", "OutputCrs-NotSupported",
    "CompressionNotSupported", "CompressionInvalid", "JpegQualityInvalid",
    "PredictorInvalid", "PredictorNotSupported", "InterleavingInvalid",
    "TilingInvalid"

))


class WCS20ExceptionHandler(Component):
    implements(ExceptionHandlerInterface)

    service = "WCS"
    versions = ("2.0.0", "2.0.1")
    request = None

    def handle_exception(self, request, exception):
        message = str(exception)
        code = getattr(exception, "code", None)
        locator = getattr(exception, "locator", None)
        status = 400

        if code is None:
            if isinstance(exception, MissingParameterException):
                code = "MissingParameterValue"
            elif isinstance(exception, DecodingException):
                code = "InvalidParameterValue"
            else:
                code = "InvalidRequest"

        if code in CODES_404:
            status = 404
        elif code in ("OperationNotSupported", "OptionNotSupported"):
            status = 501

        encoder = OWS20ExceptionXMLEncoder()
        xml = encoder.serialize(
            encoder.encode_exception(message, "2.0.1", code, locator)
        )

        return (xml, encoder.content_type, status)



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.conf import settings
from django.utils.module_loading import import_string

from eoxserver.resources.coverages.metadata.config import (
    DEFAULT_EOXS_COVERAGE_METADATA_FORMAT_READERS,
    DEFAULT_EOXS_COVERAGE_METADATA_GDAL_DATASET_FORMAT_READERS
)

METADATA_FORMAT_READERS = None
METADATA_GDAL_DATASET_FORMAT_READERS = None


def _setup_readers():
    global METADATA_FORMAT_READERS
    global METADATA_GDAL_DATASET_FORMAT_READERS
    specifiers = getattr(
        settings, 'EOXS_COVERAGE_METADATA_FORMAT_READERS',
        DEFAULT_EOXS_COVERAGE_METADATA_FORMAT_READERS
    )
    METADATA_FORMAT_READERS = [
        import_string(specifier)()
        for specifier in specifiers
    ]

    specifiers = getattr(
        settings, 'EOXS_COVERAGE_METADATA_GDAL_DATASET_FORMAT_READERS',
        DEFAULT_EOXS_COVERAGE_METADATA_GDAL_DATASET_FORMAT_READERS
    )
    METADATA_GDAL_DATASET_FORMAT_READERS = [
        import_string(specifier)()
        for specifier in specifiers
    ]


def get_reader_by_test(obj):
    """ Get a coverage metadata format reader by testing.
    """
    if not METADATA_FORMAT_READERS:
        _setup_readers()

    for reader in METADATA_FORMAT_READERS:
        if reader.test(obj):
            return reader
    return None


def get_reader_by_format(format):
    if not METADATA_FORMAT_READERS:
        _setup_readers()

    for reader in METADATA_FORMAT_READERS:
        if format in reader.formats:
            return reader
    return None


def get_gdal_dataset_format_readers():
    if not METADATA_GDAL_DATASET_FORMAT_READERS:
        _setup_readers()
    return METADATA_GDAL_DATASET_FORMAT_READERS



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------


try:
    from operator import and_, or_, add, sub, mul, div
except ImportError:
    from operator import and_, or_, add, sub, mul, truediv as div
from datetime import datetime, timedelta

try:
    from collections import OrderedDict
except ImportError:
    from django.utils.datastructures import SortedDict as OrderedDict
from django.utils.six import string_types

from django.db.models import Q, F, ForeignKey, Value
from django.db.models.expressions import Expression

from django.contrib.gis.gdal import SpatialReference
from django.contrib.gis.geos import Polygon
from django.contrib.gis.measure import D

from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import config, enum
from eoxserver.resources.coverages import models

ARITHMETIC_TYPES = (Expression, F, Value, int, float)

# ------------------------------------------------------------------------------
# Filters
# ------------------------------------------------------------------------------


def combine(sub_filters, combinator="AND"):
    """ Combine filters using a logical combinator

        :param sub_filters: the filters to combine
        :param combinator: a string: "AND" / "OR"
        :type sub_filters: ``list`` of :class:`django.db.models.Q` objects
        :return: the combined filter
        :rtype: :class:`django.db.models.Q`
    """
    for sub_filter in sub_filters:
        assert isinstance(sub_filter, Q)

    assert combinator in ("AND", "OR")
    op = and_ if combinator == "AND" else or_

    out = None
    for sub_filter in sub_filters:
        if out is None:
            out = sub_filter
        else:
            out = op(out, sub_filter)

    return out


def negate(sub_filter):
    """ Negate a filter, opposing its meaning.

        :param sub_filter: the filter to negate
        :type sub_filter: :class:`django.db.models.Q`
        :return: the negated filter
        :rtype: :class:`django.db.models.Q`
    """
    assert isinstance(sub_filter, Q)
    return ~sub_filter


OP_TO_COMP = {
    "<": "lt",
    "<=": "lte",
    ">": "gt",
    ">=": "gte",
    "<>": None,
    "=": "exact"
}


def compare(lhs, rhs, op, mapping_choices=None):
    """ Compare a filter with an expression using a comparison operation

        :param lhs: the field to compare
        :param rhs: the filter expression
        :param op: a string denoting the operation. one of ``"<"``, ``"<="``,
                   ``">"``, ``">="``, ``"<>"``, ``"="``
        :param mapping_choices: a dict to lookup potential choices for a
                                certain field.
        :type lhs: :class:`django.db.models.F`
        :type rhs: :class:`django.db.models.F`
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    assert isinstance(lhs, F)
    # assert isinstance(rhs, Q)  # TODO!!
    assert op in OP_TO_COMP
    comp = OP_TO_COMP[op]

    field_name = lhs.name

    if mapping_choices and field_name in mapping_choices:
        try:
            if isinstance(rhs, string_types):
                rhs = mapping_choices[field_name][rhs]
            elif hasattr(rhs, 'value'):
                rhs = Value(mapping_choices[field_name][rhs.value])

        except KeyError as e:
            raise AssertionError("Invalid field value %s" % e)

    if comp:
        return Q(**{"%s__%s" % (lhs.name, comp): rhs})
    return ~Q(**{field_name: rhs})


def between(lhs, low, high, not_=False):
    """ Create a filter to match elements that have a value within a certain
        range.

        :param lhs: the field to compare
        :param low: the lower value of the range
        :param high: the upper value of the range
        :param not_: whether the range shall be inclusive (the default) or
                     exclusive
        :type lhs: :class:`django.db.models.F`
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    assert isinstance(lhs, F)
    # assert isinstance(low, BaseExpression)
    # assert isinstance(high, BaseExpression)  # TODO

    q = Q(**{"%s__range" % lhs.name: (low, high)})
    return ~q if not_ else q


def like(lhs, rhs, case=False, not_=False, mapping_choices=None):
    """ Create a filter to filter elements according to a string attribute using
        wildcard expressions.

        :param lhs: the field to compare
        :param rhs: the wildcard pattern: a string containing any number of '%'
                    characters as wildcards.
        :param case: whether the lookup shall be done case sensitively or not
        :param not_: whether the range shall be inclusive (the default) or
                     exclusive
        :param mapping_choices: a dict to lookup potential choices for a
                                certain field.
        :type lhs: :class:`django.db.models.F`
        :type rhs: str
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    assert isinstance(lhs, F)

    if isinstance(rhs, string_types):
        pattern = rhs
    elif hasattr(rhs, 'value'):
        pattern = rhs.value
    else:
        raise AssertionError('Invalid pattern specified')

    parts = pattern.split("%")
    length = len(parts)

    if mapping_choices and lhs.name in mapping_choices:
        # special case when choices are given for the field:
        # compare statically and use 'in' operator to check if contained
        cmp_av = [
            (a, a if case else a.lower())
            for a in mapping_choices[lhs.name].keys()
        ]

        for idx, part in enumerate(parts):
            if not part:
                continue

            cmp_p = part if case else part.lower()

            if idx == 0 and length > 1:  # startswith
                cmp_av = [a for a in cmp_av if a[1].startswith(cmp_p)]
            elif idx == 0:  # exact matching
                cmp_av = [a for a in cmp_av if a[1] == cmp_p]
            elif idx == length - 1:   # endswith
                cmp_av = [a for a in cmp_av if a[1].endswith(cmp_p)]
            else:  # middle
                cmp_av = [a for a in cmp_av if cmp_p in a[1]]

        q = Q(**{
            "%s__in" % lhs.name: [
                mapping_choices[lhs.name][a[0]]
                for a in cmp_av
            ]
        })

    else:
        i = "" if case else "i"
        q = None

        for idx, part in enumerate(parts):
            if not part:
                continue

            if idx == 0 and length > 1:  # startswith
                new_q = Q(**{
                    "%s__%s" % (lhs.name, "%sstartswith" % i): part
                })
            elif idx == 0:  # exact matching
                new_q = Q(**{
                    "%s__%s" % (lhs.name, "%sexact" % i): part
                })
            elif idx == length - 1:   # endswith
                new_q = Q(**{
                    "%s__%s" % (lhs.name, "%sendswith" % i): part
                })
            else:  # middle
                new_q = Q(**{
                    "%s__%s" % (lhs.name, "%scontains" % i): part
                })

            q = q & new_q if q else new_q

    return ~q if not_ else q


def contains(lhs, items, not_=False, mapping_choices=None):
    """ Create a filter to match elements attribute to be in a list of choices.

        :param lhs: the field to compare
        :param items: a list of choices
        :param not_: whether the range shall be inclusive (the default) or
                     exclusive
        :param mapping_choices: a dict to lookup potential choices for a
                                certain field.
        :type lhs: :class:`django.db.models.F`
        :type items: list
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    assert isinstance(lhs, F)
    # for item in items:
    #     assert isinstance(item, BaseExpression)

    if mapping_choices and lhs.name in mapping_choices:
        def map_value(item):
            try:
                if isinstance(item, string_types):
                    item = mapping_choices[lhs.name][item]
                elif hasattr(item, 'value'):
                    item = Value(mapping_choices[lhs.name][item.value])

            except KeyError as e:
                raise AssertionError("Invalid field value %s" % e)
            return item

        items = map(map_value, items)

    q = Q(**{"%s__in" % lhs.name: items})
    return ~q if not_ else q


def null(lhs, not_=False):
    """ Create a filter to match elements whose attribute is (not) null

        :param lhs: the field to compare
        :param not_: whether the range shall be inclusive (the default) or
                     exclusive
        :type lhs: :class:`django.db.models.F`
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    assert isinstance(lhs, F)
    return Q(**{"%s__isnull" % lhs.name: not not_})


def temporal(lhs, time_or_period, op):
    """ Create a temporal filter for the given temporal attribute.

        :param lhs: the field to compare
        :param time_or_period: the time instant or time span to use as a filter
        :param op: the comparison operation. one of "BEFORE", "BEFORE OR
                   DURING", "DURING", "DURING OR AFTER", "AFTER".
        :type lhs: :class:`django.db.models.F`
        :type time_or_period: :class:`datetime.datetime` or a tuple of two
                              datetimes or a tuple of one datetime and one
                              :class:`datetime.timedelta`
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    assert isinstance(lhs, F)
    assert op in (
        "BEFORE", "BEFORE OR DURING", "DURING", "DURING OR AFTER", "AFTER"
    )
    low = None
    high = None
    if op in ("BEFORE", "AFTER"):
        assert isinstance(time_or_period, datetime)
        if op == "BEFORE":
            high = time_or_period
        else:
            low = time_or_period
    else:
        low, high = time_or_period
        assert isinstance(low, datetime) or isinstance(high, datetime)

        if isinstance(low, timedelta):
            low = high - low
        if isinstance(high, timedelta):
            high = low + high

    if low and high:
        return Q(**{"%s__range" % lhs.name: (low, high)})
    elif low:
        return Q(**{"%s__gte" % lhs.name: low})
    else:
        return Q(**{"%s__lte" % lhs.name: high})


def time_interval(time_or_period, containment='overlaps',
                  begin_time_field='begin_time', end_time_field='end_time'):
    """
    """
    config = get_eoxserver_config()
    reader = SubsetConfigReader(config)

    if reader.time_interval_interpretation == "closed":
        gt_op = "__gte"
        lt_op = "__lte"
    else:
        gt_op = "__gt"
        lt_op = "__lt"

    is_slice = len(time_or_period) == 1
    if len(time_or_period) == 1:
        is_slice = True
        value = time_or_period[0]
    else:
        is_slice = False
        low, high = time_or_period

    if is_slice or (high == low and containment == "overlaps"):
        return Q(**{
            begin_time_field + "__lte": time_or_period[0],
            end_time_field + "__gte": time_or_period[0]
        })

    elif high == low:
        return Q(**{
            begin_time_field + "__gte": value,
            end_time_field + "__lte": value
        })

    else:
        q = Q()
        # check if the temporal bounds must be strictly contained
        if containment == "contains":
            if high is not None:
                q &= Q(**{
                    end_time_field + lt_op: high
                })
            if low is not None:
                q &= Q(**{
                    begin_time_field + gt_op: low
                })
        # or just overlapping
        else:
            if high is not None:
                q &= Q(**{
                    begin_time_field + lt_op: high
                })
            if low is not None:
                q &= Q(**{
                    end_time_field + gt_op: low
                })
        return q


UNITS_LOOKUP = {
    "kilometers": "km",
    "meters": "m"
}


def spatial(lhs, rhs, op, pattern=None, distance=None, units=None):
    """ Create a spatial filter for the given spatial attribute.

        :param lhs: the field to compare
        :param rhs: the time instant or time span to use as a filter
        :param op: the comparison operation. one of "INTERSECTS", "DISJOINT",
                   "CONTAINS", "WITHIN", "TOUCHES", "CROSSES", "OVERLAPS",
                   "EQUALS", "RELATE", "DWITHIN", "BEYOND"
        :param pattern: the spatial relation pattern
        :param distance: the distance value for distance based lookups:
                         "DWITHIN" and "BEYOND"
        :param units: the units the distance is expressed in
        :type lhs: :class:`django.db.models.F`
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    assert isinstance(lhs, F)
    # assert isinstance(rhs, BaseExpression)  # TODO

    assert op in (
        "INTERSECTS", "DISJOINT", "CONTAINS", "WITHIN", "TOUCHES", "CROSSES",
        "OVERLAPS", "EQUALS", "RELATE", "DWITHIN", "BEYOND"
    )
    if op == "RELATE":
        assert pattern
    elif op in ("DWITHIN", "BEYOND"):
        assert distance
        assert units

    if op in (
            "INTERSECTS", "DISJOINT", "CONTAINS", "WITHIN", "TOUCHES",
            "CROSSES", "OVERLAPS", "EQUALS"):
        return Q(**{"%s__%s" % (lhs.name, op.lower()): rhs})
    elif op == "RELATE":
        return Q(**{"%s__relate" % lhs.name: (rhs, pattern)})
    elif op in ("DWITHIN", "BEYOND"):
        # TODO: maybe use D.unit_attname(units)
        d = D(**{UNITS_LOOKUP[units]: distance})
        if op == "DWITHIN":
            return Q(**{"%s__distance_lte" % lhs.name: (rhs, d, 'spheroid')})
        return Q(**{"%s__distance_gte" % lhs.name: (rhs, d, 'spheroid')})


def bbox(lhs, minx, miny, maxx, maxy, crs=None, bboverlaps=True):
    """ Create a bounding box filter for the given spatial attribute.

        :param lhs: the field to compare
        :param minx: the lower x part of the bbox
        :param miny: the lower y part of the bbox
        :param maxx: the upper x part of the bbox
        :param maxy: the upper y part of the bbox
        :param crs: the CRS the bbox is expressed in
        :type lhs: :class:`django.db.models.F`
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    assert isinstance(lhs, F)
    box = Polygon.from_bbox((minx, miny, maxx, maxy))

    if crs:
        box.srid = SpatialReference(crs).srid
        box.transform(4326)

    if bboverlaps:
        return Q(**{"%s__bboverlaps" % lhs.name: box})
    return Q(**{"%s__intersects" % lhs.name: box})


# ------------------------------------------------------------------------------
# Configuration
# ------------------------------------------------------------------------------

class SubsetConfigReader(config.Reader):
    section = "services.owscommon"
    time_interval_interpretation = config.Option(
        default="closed", type=enum(("closed", "open"), False)
    )

# ------------------------------------------------------------------------------
# Expressions
# ------------------------------------------------------------------------------


def attribute(name, field_mapping=None):
    """ Create an attribute lookup expression using a field mapping dictionary.

        :param name: the field filter name
        :param field_mapping: the dictionary to use as a lookup.
        :rtype: :class:`django.db.models.F`
    """
    if field_mapping:
        field = field_mapping.get(name, name)
    else:
        field = name
    return F(field)


def literal(value):
    return Value(value)


OP_TO_FUNC = {
    "+": add,
    "-": sub,
    "*": mul,
    "/": div
}


def arithmetic(lhs, rhs, op):
    """ Create an arithmetic filter

        :param lhs: left hand side of the arithmetic expression. either a
                    scalar or a field lookup or another type of expression
        :param rhs: same as `lhs`
        :param op: the arithmetic operation. one of "+", "-", "*", "/"
        :rtype: :class:`django.db.models.F`
    """

    assert isinstance(lhs, ARITHMETIC_TYPES), (
        '%r is not a compatible type' % lhs
    )
    assert isinstance(rhs, ARITHMETIC_TYPES), (
        '%r is not a compatible type' % rhs
    )
    assert op in OP_TO_FUNC
    func = OP_TO_FUNC[op]
    return func(lhs, rhs)


def get_field_mapping_for_model(model_class, strict=False):
    """ Utility function to get the metadata mapping for a specific model class.

        :param model_class: The django database model to create the mapping for
        :param strict: Whether only the related metadata attributes shall be
                       included or the basic ones as-well
        :returns: two dictionaries: the mapping dict, mapping from metadata
                  filter name to the database field lookup and a dict to map
                  the field lookup to the potential choices.
    """
    mapping = OrderedDict()
    mapping_choices = {}

    metadata_classes = {
        models.Collection: (models.CollectionMetadata, 'collection'),
        models.Product: (models.ProductMetadata, 'product'),
        models.Coverage: (models.CoverageMetadata, 'coverage'),
    }

    if issubclass(model_class, models.EOObject) and not strict:
        field_names = ('identifier', 'begin_time', 'end_time', 'footprint')
        for field_name in field_names:
            mapping[_to_camel_case(field_name)] = field_name

    if model_class in metadata_classes:
        new_mapping, mapping_choices = _get_metadata_model_mapping(
            *metadata_classes.get(model_class)
        )
        mapping.update(new_mapping)

    elif model_class is models.EOObject:
        for metadata_class, name in metadata_classes.values():
            class_mapping, class_choices = _get_metadata_model_mapping(
                metadata_class, "%s__%s" % (name, name)
            )
            mapping.update(class_mapping)
            mapping_choices.update(class_choices)

    return mapping, mapping_choices


# helpers
def _to_camel_case(word):
    string = ''.join(x.capitalize() or '_' for x in word.split('_'))
    return string[0].lower() + string[1:]


def _is_common_value(field):
    try:
        if isinstance(field, ForeignKey):
            field.related_model._meta.get_field('value')
            return True
    except Exception:
        pass
    return False


def _get_metadata_model_mapping(metadata_class, path_name):
    mapping = OrderedDict()
    mapping_choices = {}
    for field in metadata_class._meta.fields:
        # skip fields that are defined in a parent model
        if field.model is not metadata_class or field.name == "id":
            continue
        if _is_common_value(field):
            full_path = '%s_metadata__%s__value' % (
                path_name, field.name
            )
        else:
            full_path = '%s_metadata__%s' % (path_name, field.name)
        mapping[_to_camel_case(field.name)] = full_path
        if field.choices:
            mapping_choices[full_path] = dict(
                (full, abbrev) for (abbrev, full) in field.choices
            )

    return mapping, mapping_choices



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2012 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
from django.utils.six import string_types
# global format selection registry
_registry = {}

def get_format_selection(driver, *args, **kwargs):
    """ Format selection factory method. """

    try:
        return _registry[driver](*args, **kwargs)
    except KeyError:
        raise ValueError("No format selection available for driver '%s'" % driver)


class FormatSelectionMetaclass(type):
    """ Metaclass for format selections
    """
    def __init__(cls, name, bases, dct):
        if isinstance(dct["driver_name"], string_types):
            _registry[dct["driver_name"]] = cls
        super(FormatSelectionMetaclass, cls).__init__(name, bases, dct)


#===============================================================================
# Format selection
#===============================================================================

class FormatSelection(object):
    """ Format selection with format specific options. """

    __metaclass__ = FormatSelectionMetaclass


    @property
    def driver_name(self):
        raise NotImplementedError


    @property
    def extension(self):
        raise NotImplementedError


    @property
    def creation_options(self):
        return []



class GeoTIFFFormatSelection(FormatSelection):
    """ Format selection for GeoTIFF format.
    """

    SUPPORTED_COMPRESSIONS = ("JPEG", "LZW", "PACKBITS", "DEFLATE", "CCITTRLE",
                              "CCITTFAX3", "CCITTFAX4", "NONE")


    def __init__(self, tiling=True, compression=None,
                 jpeg_quality=None, zlevel=None, creation_options=None):
        self.final_options = {}
        if compression:
            compression = compression.upper()
            if compression not in self.SUPPORTED_COMPRESSIONS:
                raise Exception("Unsupported compression method. Supported "
                                "compressions are: %s"
                                % ", ".join(self.SUPPORTED_COMPRESSIONS))
            self.final_options["COMPRESS"] = compression

            if jpeg_quality is not None and compression == "JPEG":
                self.final_options["JPEG_QUALITY"] = jpeg_quality
            elif jpeg_quality is not None:
                raise ValueError("'jpeg_quality' can only be used with JPEG "
                                 "compression")

            if zlevel is not None and compression == "DEFLATE":
                self.final_options["ZLEVEL"] = zlevel
            elif zlevel is not None:
                raise ValueError("'zlevel' can only be used with DEFLATE "
                                 "compression")

        if tiling:
            self.final_options["TILED"] = "YES"

        if creation_options:
            self.final_options.update(dict(creation_options))


    driver_name = "GTiff"
    extension = ".tif"


    @property
    def creation_options(self):
        return ["%s=%s" % (key, value)
                for key, value in self.final_options.items()]




#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2012 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import logging

from eoxserver.contrib import gdal, ogr, osr
from eoxserver.processing.gdal import reftools as rt
from eoxserver.processing.preprocessing.util import (
    create_mem, copy_metadata
)
from eoxserver.processing.preprocessing.exceptions import GCPTransformException


logger = logging.getLogger(__name__)


#===============================================================================
# Geographic references
#===============================================================================

class GeographicReference(object):
    pass


class Extent(GeographicReference):
    """ Sets the extent of the dataset expressed as a 4-tuple (minx, miny, maxx,
        maxy) with an optional SRID (defaults to EPSG: 4326).
    """
    def __init__(self, minx, miny, maxx, maxy, srid=4326):
        self.minx, self.miny, self.maxx, self.maxy = minx, miny, maxx, maxy
        self.srid = srid


    def apply(self, ds):
        """ Set the geotransform and projection of the dataset according to
            the defined extent and SRID.
        """
        sr = osr.SpatialReference(); sr.ImportFromEPSG(self.srid)
        ds.SetGeoTransform([
            self.minx,
            (self.maxx - self.minx) / ds.RasterXSize,
            0,
            self.maxy,
            0,
            -(self.maxy - self.miny) / ds.RasterYSize
        ])
        ds.SetProjection(sr.ExportToWkt())

        return ds, None


class GCPList(GeographicReference):
    """ Sets a list of GCPs (Ground Control Points) to the dataset and then
        performs a rectification to a projection specified by SRID.
    """

    def __init__(self, gcps, gcp_srid=4326, srid=None):
        """ Expects a list of GCPs as a list of tuples in the form
            'x,y,[z,]pixel,line'.
        """

        self.gcps = map(lambda gcp: gdal.GCP(*gcp) if len(gcp) == 5
                        else gdal.GCP(gcp[0], gcp[1], 0.0, gcp[2], gcp[3]),
                        gcps)
        self.gcp_srid = gcp_srid
        self.srid = srid


    def apply(self, src_ds):
        # setup
        dst_sr = osr.SpatialReference()
        gcp_sr = osr.SpatialReference()

        dst_sr.ImportFromEPSG(self.srid if self.srid is not None
                              else self.gcp_srid)
        gcp_sr.ImportFromEPSG(self.gcp_srid)


        logger.debug("Using GCP Projection '%s'" % gcp_sr.ExportToWkt())
        logger.debug("Applying GCPs: MULTIPOINT(%s) -> MULTIPOINT(%s)"
                      % (", ".join([("(%f %f)") % (gcp.GCPX, gcp.GCPY) for gcp in self.gcps]) ,
                      ", ".join([("(%f %f)") % (gcp.GCPPixel, gcp.GCPLine) for gcp in self.gcps])))
        # set the GCPs
        src_ds.SetGCPs(self.gcps, gcp_sr.ExportToWkt())

        # Try to find and use the best transform method/order.
        # Orders are: -1 (TPS), 3, 2, and 1 (all GCP)
        # Loop over the min and max GCP number to order map.
        for min_gcpnum, max_gcpnum, order in [(3, None, -1), (10, None, 3), (6, None, 2), (3, None, 1)]:
            # if the number of GCP matches
            if len(self.gcps) >= min_gcpnum and (max_gcpnum is None or len(self.gcps) <= max_gcpnum):
                try:

                    if ( order < 0 ) :
                        # let the reftools suggest the right interpolator
                        rt_prm = rt.suggest_transformer( src_ds )
                    else:
                        # use the polynomial GCP interpolation as requested
                        rt_prm = { "method":rt.METHOD_GCP, "order":order }

                    logger.debug("Trying order '%i' {method:%s,order:%s}" % \
                        (order, rt.METHOD2STR[rt_prm["method"]] , rt_prm["order"] ) )
                    # get the suggested pixel size/geotransform
                    size_x, size_y, geotransform = rt.suggested_warp_output(
                        src_ds,
                        None,
                        dst_sr.ExportToWkt(),
                        **rt_prm
                    )
                    if size_x > 100000 or size_y > 100000:
                        raise RuntimeError("Calculated size exceeds limit.")
                    logger.debug("New size is '%i x %i'" % (size_x, size_y))

                    # create the output dataset
                    dst_ds = create_mem(size_x, size_y,
                                        src_ds.RasterCount,
                                        src_ds.GetRasterBand(1).DataType)

                    # reproject the image
                    dst_ds.SetProjection(dst_sr.ExportToWkt())
                    dst_ds.SetGeoTransform(geotransform)

                    rt.reproject_image(src_ds, "", dst_ds, "", **rt_prm )

                    copy_metadata(src_ds, dst_ds)

                    # retrieve the footprint from the given GCPs
                    footprint_wkt = rt.get_footprint_wkt(src_ds, **rt_prm )

                except RuntimeError as e:
                    logger.debug("Failed using order '%i'. Error was '%s'."
                                 % (order, str(e)))
                    # the given method was not applicable, use the next one
                    continue

                else:
                    logger.debug("Successfully used order '%i'" % order)
                    # the transform method was successful, exit the loop
                    break
        else:
            # no method worked, so raise an error
            raise GCPTransformException("Could not find a valid transform method.")

        # reproject the footprint to a lon/lat projection if necessary
        if not gcp_sr.IsGeographic():
            out_sr = osr.SpatialReference()
            out_sr.ImportFromEPSG(4326)
            geom = ogr.CreateGeometryFromWkt(footprint_wkt, gcp_sr)
            geom.TransformTo(out_sr)
            footprint_wkt = geom.ExportToWkt()

        logger.debug("Calculated footprint: '%s'." % footprint_wkt)

        return dst_ds, footprint_wkt




# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

try:
    from itertools import izip_longest
except ImportError:
    from itertools import zip_longest as izip_longest

from copy import deepcopy

from django.utils.six import string_types
from eoxserver.core.util.timetools import parse_iso8601, parse_duration
from eoxserver.contrib import gdal, osr
from eoxserver.contrib.osr import SpatialReference
from eoxserver.backends.access import get_vsi_path, get_vsi_env

GRID_TYPE_ELEVATION = 1
GRID_TYPE_TEMPORAL = 2


def is_referenceable(grid_model):
    return grid_model.axis_1_offset is None


class Field(object):
    def __init__(self, index, identifier, description, definition,
                 unit_of_measure, wavelength, significant_figures,
                 allowed_values, nil_values, data_type, data_type_range):
        self._index = index
        self._identifier = identifier
        self._description = description
        self._definition = definition
        self._unit_of_measure = unit_of_measure
        self._wavelength = wavelength
        self._significant_figures = significant_figures
        self._allowed_values = allowed_values
        self._nil_values = nil_values
        self._data_type = data_type
        self._data_type_range = data_type_range

    @property
    def index(self):
        return self._index

    @property
    def identifier(self):
        return self._identifier

    @property
    def description(self):
        return self._description

    @property
    def definition(self):
        return self._definition

    @property
    def unit_of_measure(self):
        return self._unit_of_measure

    @property
    def wavelength(self):
        return self._wavelength

    @property
    def significant_figures(self):
        return self._significant_figures

    @property
    def allowed_values(self):
        return self._allowed_values

    @property
    def nil_values(self):
        return self._nil_values

    @property
    def data_type(self):
        return self._data_type

    @property
    def data_type_range(self):
        return self._data_type_range

    def __eq__(self, other):
        try:
            return (
                self._identifier == other._identifier and
                self._description == other._description and
                self._definition == other._definition and
                self._unit_of_measure == other._unit_of_measure and
                self._wavelength == other._wavelength and
                self._significant_figures == other._significant_figures and
                self._allowed_values == other._allowed_values and
                self._nil_values == other._nil_values and
                self._data_type == other._data_type and
                self._data_type_range == other._data_type_range
            )
        except AttributeError:
            return False

    def __repr__(self):
        return '<Field %r>' % self.identifier


class RangeType(list):
    def __init__(self, name, fields):
        super(RangeType, self).__init__(fields)
        self._name = name

    @property
    def name(self):
        return self._name

    def get_field(self, name):
        try:
            return next(
                field
                for field in self
                if field.identifier == name
            )
        except StopIteration:
            raise KeyError(name)

    def subset(self, subsets):
        fields = []
        for subset in subsets:
            if isinstance(subset, string_types):
                fields.append(deepcopy(self.get_field(subset)))
            elif isinstance(subset, (list, tuple)):
                start_id, stop_id = subset
                start = self.index(self.get_field(start_id))
                stop = self.index(self.get_field(stop_id))
                fields.extend([
                    deepcopy(item)
                    for item in self[start:stop + (1 if stop > start else -1)]
                ])

        for i, field in enumerate(fields):
            field._index = i

        return type(self)(self.name, fields)

    @classmethod
    def from_coverage_type(cls, coverage_type):
        def get_data_type(field_type):
            numbits = (
                field_type.numbits if field_type.numbits is not None else 16
            )
            signed = field_type.signed
            is_float = field_type.is_float

            if is_float:
                if numbits <= 32:
                    return gdal.GDT_Float32
                return gdal.GDT_Float64
            elif signed:
                if numbits <= 8:
                    return gdal.GDT_Byte
                elif numbits <= 16:
                    return gdal.GDT_Int16
                else:
                    return gdal.GDT_Int32
            else:
                if numbits <= 8:
                    return gdal.GDT_Byte
                elif numbits <= 16:
                    return gdal.GDT_UInt16
                else:
                    return gdal.GDT_UInt32
            return gdal.GDT_Unknown

        def get_data_type_range(field_type):
            numbits = (
                field_type.numbits if field_type.numbits is not None else 32
            )
            signed = field_type.signed
            is_float = field_type.is_float
            if is_float:
                if numbits == 32:
                    return gdal.GDT_NUMERIC_LIMITS[gdal.GDT_Float32]
                elif numbits == 64:
                    return gdal.GDT_NUMERIC_LIMITS[gdal.GDT_Float64]
            elif signed:
                max_ = 2 ** (numbits - 1)
                return (-max_, max_ - 1)
            else:
                return (0, 2 ** numbits)

        return cls(coverage_type.name, [
            Field(
                index=i,
                identifier=field_type.identifier,
                description=field_type.description,
                definition=field_type.definition,
                unit_of_measure=field_type.unit_of_measure,
                wavelength=field_type.wavelength,
                significant_figures=field_type.significant_figures,
                allowed_values=[
                    (value_range.start, value_range.end)
                    for value_range in field_type.allowed_value_ranges.all()
                ],
                nil_values=[
                    (nil_value.value, nil_value.reason)
                    for nil_value in field_type.nil_values.all()
                ],
                data_type=get_data_type(field_type),
                data_type_range=get_data_type_range(field_type)
            )
            for i, field_type in enumerate(coverage_type.field_types.all())
        ])

    @classmethod
    def from_gdal_dataset(cls, ds, base_identifier):
        fields = []
        bandoffset = 0
        for i in range(ds.RasterCount):
            band = ds.GetRasterBand(i + 1)
            nodata_value = band.GetNoDataValue()
            if nodata_value is not None:
                nil_values = [(nodata_value, "")]
            else:
                nil_values = []

            fields.append(
                Field(
                    index=i,
                    identifier="%s_%d" % (base_identifier, bandoffset + i),
                    # TODO: get info from band metadata?
                    description="",
                    definition="",
                    unit_of_measure="",
                    wavelength="",
                    significant_figures=gdal.GDT_SIGNIFICANT_FIGURES.get(
                        band.DataType
                    ),
                    allowed_values=[
                        gdal.GDT_NUMERIC_LIMITS[band.DataType]
                    ]
                    if band.DataType in gdal.GDT_NUMERIC_LIMITS else [],
                    nil_values=nil_values,
                    data_type=band.DataType,
                    data_type_range=gdal.GDT_NUMERIC_LIMITS.get(band.DataType)
                )
            )
            bandoffset += 1
        return cls(base_identifier, fields)


class Axis(object):
    regular = True

    def __init__(self, name, type, offset, uom=None):
        self._name = name
        self._type = type
        self._offset = offset
        self._uom = uom

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    @property
    def offset(self):
        return self._offset

    @property
    def uom(self):
        return self._uom


class IrregularAxis(object):
    regular = False

    def __init__(self, name, type, positions, uom=None):
        self._name = name
        self._type = type
        self._positions = positions
        self._uom = uom

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    @property
    def positions(self):
        return self._positions

    @property
    def uom(self):
        return self._uom


class Grid(list):
    def __init__(self, coordinate_reference_system, axes):
        super(Grid, self).__init__(axes)
        self._coordinate_reference_system = coordinate_reference_system

    @classmethod
    def from_model(cls, grid_model):
        is_ref = is_referenceable(grid_model)
        names = grid_model.axis_names
        types = grid_model.axis_types
        offsets = grid_model.axis_offsets

        axes = []

        axes_iter = izip_longest(names, types, offsets)
        for name, type_, offset in axes_iter:
            if is_ref:
                offset = None
            elif type_ == GRID_TYPE_TEMPORAL:
                offset = parse_duration(offset)
            else:
                offset = float(offset)

            axes.append(Axis(name, type_, offset))

        return cls(grid_model.coordinate_reference_system, axes)

    @classmethod
    def from_gdal_dataset(cls, ds):
        projection = ds.GetProjection()
        gt = ds.GetGeoTransform()
        sr = osr.SpatialReference(projection)

        axis_names = ['x', 'y'] if sr.IsProjected() else ['long', 'lat']

        return cls(projection, [
            Axis(axis_names[0], 'spatial', gt[1]),
            Axis(axis_names[1], 'spatial', gt[5]),
        ])

    @property
    def spatial_reference(self):
        return SpatialReference(self.coordinate_reference_system)

    @property
    def coordinate_reference_system(self):
        return self._coordinate_reference_system

    @property
    def names(self):
        return [axis.name for axis in self]

    @property
    def types(self):
        return [axis.type for axis in self]

    @property
    def offsets(self):
        return [axis.offset for axis in self]

    @property
    def has_elevation(self):
        return GRID_TYPE_ELEVATION in self.types

    @property
    def has_temporal(self):
        return GRID_TYPE_TEMPORAL in self.types

    @property
    def is_referenceable(self):
        return self[0].offset is None


class Origin(list):
    @classmethod
    def from_description(cls, axis_types, origins):
        return cls([
            parse_iso8601(orig) if type_ == GRID_TYPE_TEMPORAL else float(orig)
            for type_, orig in zip(axis_types, origins)
        ])

    @classmethod
    def from_gdal_dataset(cls, ds):
        gt = ds.GetGeoTransform()
        return cls([
            float(gt[0]),
            float(gt[3]),
        ])


class EOMetadata(object):
    def __init__(self, begin_time, end_time, footprint):
        self._begin_time = begin_time
        self._end_time = end_time
        self._footprint = footprint

    @property
    def footprint(self):
        return self._footprint

    @property
    def begin_time(self):
        return self._begin_time

    @property
    def end_time(self):
        return self._end_time


class Location(object):
    def __init__(self, path, env, format):
        self._path = path
        self._env = env
        self._format = format

    @property
    def path(self):
        return self._path

    @property
    def env(self):
        return self._env

    @property
    def format(self):
        return self._format


class ArraydataLocation(Location):
    def __init__(self, path, env, format, start_field, end_field):
        super(ArraydataLocation, self).__init__(path, env, format)
        self._start_field = start_field
        self._end_field = end_field

    @property
    def start_field(self):
        return self._start_field

    @property
    def end_field(self):
        return self._end_field

    @property
    def field_count(self):
        return self._end_field - self._start_field + 1

    def field_index_to_band_index(self, field_index):
        return field_index - self.start_field


class Coverage(object):
    """ Representation of a coverage for internal processing.
    """
    def __init__(self, identifier, eo_metadata, range_type, grid, origin, size,
                 arraydata_locations, metadata_locations, native_format=None):
        self._identifier = identifier
        self._eo_metadata = eo_metadata
        self._range_type = range_type
        self._origin = origin
        self._grid = grid
        self._size = size
        self._arraydata_locations = arraydata_locations
        self._metadata_locations = metadata_locations
        self._native_format = native_format

    @property
    def identifier(self):
        return self._identifier

    @property
    def eo_metadata(self):
        return self._eo_metadata

    @property
    def footprint(self):
        return self._eo_metadata.footprint if self._eo_metadata else None

    @property
    def begin_time(self):
        return self._eo_metadata.begin_time if self._eo_metadata else None

    @property
    def end_time(self):
        return self._eo_metadata.end_time if self._eo_metadata else None

    @property
    def range_type(self):
        return self._range_type

    @property
    def origin(self):
        return self._origin

    @property
    def grid(self):
        return self._grid

    @property
    def size(self):
        return tuple(self._size)

    @property
    def native_format(self):
        return self._native_format or (
            self.arraydata_locations[0].format
            if self.arraydata_locations else
            None
        )

    @property
    def arraydata_locations(self):
        return self._arraydata_locations

    @property
    def metadata_locations(self):
        return self._metadata_locations

    @property
    def coverage_subtype(self):
        subtype = "RectifiedDataset"
        if not self.footprint or not self.begin_time or not self.end_time:
            subtype = "RectifiedGridCoverage"
        elif self.grid.is_referenceable:
            subtype = "ReferenceableDataset"
        return subtype

    @property
    def extent(self):
        types = self.grid.types
        offsets = self.grid.offsets

        lows = []
        highs = []

        if offsets[0] is not None:
            axes = izip_longest(types, offsets, self.origin, self.size)
            for type_, offset, origin, size in axes:
                a = origin
                b = origin + size * offset

                if offset > 0:
                    lows.append(a)
                    highs.append(b)
                else:
                    lows.append(b)
                    highs.append(a)

            return tuple(lows + highs)
        elif self.footprint:
            return self.footprint.extent

    def get_location_for_field(self, field_or_identifier):
        if isinstance(field_or_identifier, Field):
            field = field_or_identifier
            if field not in self.range_type:
                return None
        else:
            try:
                field = next(
                    field
                    for field in self.range_type
                    if field.identifier == field_or_identifier
                )
            except StopIteration:
                return None

        index = field.index
        for location in self.arraydata_locations:
            if index >= location.start_field and index <= location.end_field:
                return location

    def get_band_index_for_field(self, field_or_identifier):
        if isinstance(field_or_identifier, Field):
            field = field_or_identifier
            if field not in self.range_type:
                return None
        else:
            try:
                field = next(
                    field
                    for field in self.range_type
                    if field.identifier == field_or_identifier
                )
            except StopIteration:
                return None

        index = field.index
        for location in self.arraydata_locations:
            if index >= location.start_field and index <= location.end_field:
                return index - location.start_field + 1

    @classmethod
    def from_model(cls, model):
        # use coverages EO metadata by default and fill up with
        # EO metadata from Product
        begin_time = model.begin_time
        if not begin_time and model.parent_product:
            begin_time = model.parent_product.begin_time
        end_time = model.end_time
        if not end_time and model.parent_product:
            end_time = model.parent_product.end_time
        footprint = model.footprint
        if not footprint and model.parent_product:
            footprint = model.parent_product.footprint
        eo_metadata = EOMetadata(begin_time, end_time, footprint)

        arraydata_locations = [
            ArraydataLocation(
                get_vsi_path(item), get_vsi_env(item.storage), item.format,
                item.field_index, item.field_index + (item.band_count - 1)
            )
            for item in model.arraydata_items.all()
        ]

        metadata_locations = [
            Location(
                get_vsi_path(item), get_vsi_env(item.storage), item.format
            )
            for item in model.metadata_items.all()
        ]

        if model.coverage_type:
            range_type = RangeType.from_coverage_type(
                model.coverage_type
            )
        else:
            range_type = RangeType.from_gdal_dataset(
                gdal.OpenShared(arraydata_locations[0].path),
                model.identifier
            )

        grid = Grid.from_model(model.grid)

        origin = Origin.from_description(grid.types, model.origin)

        return cls(
            identifier=model.identifier,
            eo_metadata=eo_metadata,
            range_type=range_type,
            origin=origin,
            grid=grid,
            size=model.size,
            arraydata_locations=arraydata_locations,
            metadata_locations=metadata_locations
        )


class Mosaic(object):
    def __init__(self, identifier, eo_metadata, range_type, grid, origin, size,
                 coverages=None):
        self._identifier = identifier
        self._eo_metadata = eo_metadata
        self._range_type = range_type
        self._origin = origin
        self._grid = grid
        self._size = size
        self._coverages = coverages if coverages is not None else []

    @property
    def identifier(self):
        return self._identifier

    @property
    def eo_metadata(self):
        return self._eo_metadata

    @property
    def footprint(self):
        return self._eo_metadata.footprint if self._eo_metadata else None

    @property
    def begin_time(self):
        return self._eo_metadata.begin_time if self._eo_metadata else None

    @property
    def end_time(self):
        return self._eo_metadata.end_time if self._eo_metadata else None

    @property
    def range_type(self):
        return self._range_type

    @property
    def origin(self):
        return self._origin

    @property
    def grid(self):
        return self._grid

    @property
    def size(self):
        return tuple(self._size)

    # @property
    # def coverage_subtype(self):
    #     subtype = "DatasetSeries"
    #     if not self.footprint or not self.begin_time or not self.end_time:
    #         subtype = "RectifiedStitchedMosaic"
    #     elif self.grid.is_referenceable:
    #         subtype = "ReferenceableStitchedMosaic"
    #     return subtype

    @property
    def extent(self):
        if not self.grid:
            return None

        types = self.grid.types
        offsets = self.grid.offsets

        lows = []
        highs = []

        axes = izip_longest(types, offsets, self.origin, self.size)
        for type_, offset, origin, size in axes:
            a = origin
            b = origin + size * offset

            if offset > 0:
                lows.append(a)
                highs.append(b)
            else:
                lows.append(b)
                highs.append(a)

        return tuple(lows + highs)

    @property
    def coverages(self):
        return self._coverages

    @classmethod
    def from_model(cls, mosaic_model, coverage_models=None):
        eo_metadata = EOMetadata(None, None, None)
        if mosaic_model.begin_time and mosaic_model.end_time and \
                mosaic_model.footprint:
            eo_metadata = EOMetadata(
                mosaic_model.begin_time, mosaic_model.end_time,
                mosaic_model.footprint
            )

        range_type = RangeType.from_coverage_type(
            mosaic_model.coverage_type
        )

        grid_model = mosaic_model.grid
        grid = None
        origin = None
        if grid_model:
            grid = Grid.from_model(grid_model)
            origin = Origin.from_description(grid.types, mosaic_model.origin)

        coverages = [
            Coverage.from_model(coverage_model)
            for coverage_model in coverage_models
        ] if coverage_models is not None else None

        return cls(
            identifier=mosaic_model.identifier,
            eo_metadata=eo_metadata, range_type=range_type, origin=origin,
            grid=grid, size=mosaic_model.size, coverages=coverages
        )


class DatasetSeries(object):
    def __init__(self, identifier, footprint=None,
                 begin_time=None, end_time=None):
        self._identifier = identifier
        self._footprint = footprint
        self._begin_time = begin_time
        self._end_time = end_time

    @property
    def identifier(self):
        return self._identifier

    @property
    def footprint(self):
        return self._footprint

    @property
    def begin_time(self):
        return self._begin_time

    @property
    def end_time(self):
        return self._end_time

    @classmethod
    def from_model(cls, model):
        return cls(
            model.identifier, model.footprint, model.begin_time, model.end_time
        )


def from_model(eo_object_model):
    from eoxserver.resources.coverages import models

    if isinstance(eo_object_model, models.Coverage):
        return Coverage.from_model(eo_object_model)
    elif isinstance(eo_object_model, models.Mosaic):
        return Mosaic.from_model(eo_object_model)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import os.path
from glob import glob
import logging
from unittest import skip

from django.test import TestCase

from eoxserver.backends import testbase
from eoxserver.backends import models
from eoxserver.backends.cache import CacheContext
from eoxserver.backends.access import retrieve
from eoxserver.backends.component import BackendComponent, env
from eoxserver.backends.testbase import withFTPServer


logger = logging.getLogger(__name__)


""" "New" data models
"""


def create(Model, *args, **kwargs):
    model = Model(*args, **kwargs)
    model.full_clean()
    model.save()
    return model


# class RetrieveTestCase(TestCase):
#     def setUp(self):
#         pass

#     def tearDown(self):
#         pass

#     def test_retrieve_http(self):
#         import storages, packages

#         storage = create(models.Storage,
#             url="http://eoxserver.org/export/2523/downloads",
#             storage_type="HTTP"
#         )
#         dataset = create(models.DataItem,
#             location="EOxServer_documentation-0.3.0.pdf",
#             storage=storage,
#             semantic="pdffile"
#         )

#         with CacheContext() as c:
#             cache_path = retrieve(dataset, c)
#             self.assertTrue(os.path.exists(cache_path))

#         self.assertFalse(os.path.exists(cache_path))

#     @skip("not yet implemented")
#     @withFTPServer()
#     def test_retrieve_ftp_zip(self):
#         import storages
#         import packages

#         storage = create(models.Storage,
#             url="ftp://anonymous:@localhost:2121/",
#             storage_type="FTP"
#         )

#         package = create(models.Package,
#             location="package.zip",
#             format="ZIP",
#             storage=storage
#         )

#         dataset = create(models.DataItem,
#             location="file.txt",
#             package=package,
#             semantic="textfile"
#         )

#         dataset2 = create(models.DataItem,
#             location="file2.txt",
#             package=package,
#             semantic="textfile"
#         )

#         with CacheContext() as c:
#             cache_path = retrieve(dataset, c)
#             cache_path2 = retrieve(dataset2, c)
#             self.assertTrue(os.path.exists(cache_path))
#             self.assertTrue(os.path.exists(cache_path2))

#             with open(cache_path) as f:
#                 self.assertEqual(f.read(), "test\n")

#             with open(cache_path2) as f:
#                 self.assertEqual(f.read(), "test 2\n")

#         self.assertFalse(os.path.exists(cache_path))
#         self.assertFalse(os.path.exists(cache_path2))



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import re
from datetime import datetime
from os.path import splitext

from django.utils.timezone import utc


class GDALDatasetEnvisatMetadataFormatReader(object):
    """ Metadata format reader for specific ENVISAT products.
    """

    def test_ds(self, ds):
        """ Check whether or not the dataset seems to be an ENVISAT image and
            has the correct metadata tags.
        """
        md_dict = ds.GetMetadata_Dict()
        for key in ("MPH_PRODUCT", "MPH_SENSING_START", "MPH_SENSING_STOP"):
            if key not in md_dict:
                return False
        if ds.GetGCPCount() == 0:
            return False

        return True

    def read_ds(self, ds):
        """ Return the ENVISAT specific metadata items.
        """
        return {
            "identifier": splitext(ds.GetMetadataItem("MPH_PRODUCT"))[0],
            "begin_time": parse_datetime(ds.GetMetadataItem("MPH_SENSING_START")),
            "end_time": parse_datetime(ds.GetMetadataItem("MPH_SENSING_STOP"))
        }


MONTHS = {
    "JAN": 1,
    "FEB": 2,
    "MAR": 3,
    "APR": 4,
    "MAY": 5,
    "JUN": 6,
    "JUL": 7,
    "AUG": 8,
    "SEP": 9,
    "OCT": 10,
    "NOV": 11,
    "DEC": 12
}


def parse_datetime(timestamp):
    """ Datetime parsing function for special Envisat datetime format.
    """
    
    
    match = re.match(
        r"(\d{2})-([A-Z]{3})-(\d{4}) (\d{2}):(\d{2}):(\d{2}).*", timestamp
    )
    day = int(match.group(1))
    month = MONTHS[match.group(2)]
    year = int(match.group(3))
    hour = int(match.group(4))
    minute = int(match.group(5))
    second = int(match.group(6))
    
    return datetime(year, month, day, hour, minute, second, tzinfo=utc)



#-------------------------------------------------------------------------------
#
#  WPS Literal Data type
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

try:
    # available in Python 2.7+
    from collections import OrderedDict
except ImportError:
    from django.utils.datastructures import SortedDict as OrderedDict

from .base import Parameter
from .data_types import BaseType, String, DTYPES
from .allowed_values import BaseAllowed, AllowedAny, AllowedEnum
from .units import UnitOfMeasure, UnitLinear
from django.utils.encoding import smart_text
from django.utils.six import text_type


class LiteralData(Parameter):
    """ Literal-data parameter class.

    Constructor parameters:
        identifier  identifier of the parameter used by the WPS service
        title       optional human-readable name (defaults to identifier)
        abstract    optional human-readable verbose description
        metadata    optional metadata (title/URL dictionary)
        optional    optional boolean flag indicating whether the input
                    parameter is optional or not
        dtype       optional data type of the parameter. String type
                    ``str`` is set by default. For list of supported
                    types see ``LiteralData.SUPPORTED_TYPES``)
        uoms        optional sequence of the supported units
        default     optional default input value. Presence of the
                    default value sets the parameter optional.
        allowed_values optional restriction on the accepted values.
                    By default any value of the given type is
                    supported. The allowed value can be specified by
                    an enumerated list (iterable) of values or by
                    instance of one of the following classes:
                    ``AllowedAny``, ``AllowedEnum``, ``AllowedRange``,
                    or ``AllowedByReference``.
        resolve_input_references  Set this option to False not to resolve
                    input references. By default the references are
                    resolved (downloaded and parsed) transparently.
                    If set to False the references must be handled
                    by the process.
    """

    def __init__(self, identifier, dtype=String, uoms=None, default=None,
                 allowed_values=None, *args, **kwargs):
        # pylint: disable=too-many-arguments, too-many-branches
        super(LiteralData, self).__init__(identifier, *args, **kwargs)

        if isinstance(dtype, type) and issubclass(dtype, BaseType):
            self._dtype = dtype
        elif isinstance(dtype, BaseType):
            self._dtype = dtype
        elif dtype in DTYPES:
            self._dtype = DTYPES[dtype]
        else:
            raise TypeError("Non-supported data type %s!" % dtype)

        if isinstance(allowed_values, BaseAllowed):
            if (hasattr(allowed_values, 'dtype') and
                    self._dtype != allowed_values.dtype):
                raise TypeError(
                    "The allowed values has a different data-type "
                    "then the literal data object %s != %s" %
                    (allowed_values.dtype, self._dtype)
                )
            self._allowed_values = allowed_values
        elif allowed_values is not None:
            self._allowed_values = AllowedEnum(allowed_values, self._dtype)
        else:
            self._allowed_values = AllowedAny()  # pylint: disable=redefined-variable-type

        if uoms:  # the first UOM is the default one
            tmp = OrderedDict()
            for uom in uoms:
                if not isinstance(uom, UnitOfMeasure):
                    uom = UnitLinear(uom[0], uom[1])
                tmp[uom.name] = uom
            self._uoms = tmp
        else:
            self._uoms = None

        if default is None:
            self.default = None
        else:
            self.default = self.parse(default)
            self.is_optional = True

    @property
    def default_uom(self):
        """ Get the default UOM. """
        return list(self._uoms)[0] if self._uoms else None

    @property
    def uoms(self):
        """ Get all allowed UOMs. """
        return list(self._uoms) if self._uoms else None

    @property
    def dtype(self):
        """Data type class of the literal data object. (RO)"""
        return self._dtype

    @property
    def allowed_values(self):
        """Allowed values object of the literal data object. (RO)"""
        return self._allowed_values

    def check(self, value):
        """Check whether the value is allowed (True) or not (False)."""
        return self._allowed_values.check(value)

    def verify(self, value):
        """Return the value if allowed or raise the ValueError exception."""
        return self._allowed_values.verify(value)

    def apply_uom(self, value, uom):
        """ Convert value from the common base to the desired UOM. """
        if uom is None:
            return value
        try:
            return self._uoms[uom].apply(value)
        except KeyError:
            raise ValueError("Invalid UOM '%s'!" % uom)

    def strip_uom(self, value, uom):
        """ Convert value from the provided UOM to the common base. """
        if uom is None:
            return value
        try:
            return self._uoms[uom].strip(value)
        except KeyError:
            raise ValueError("Invalid UOM '%s'!" % uom)

    def encode(self, value, uom=None, encoding=None):
        """ Encode the output value to its string representation.

            The value is checked to match the defined allowed values
            restriction and the UOM conversion is applied.

            Returns Unicode or byte-string if the encoding is given.
        """
        try:
            _value = self._allowed_values.verify(value)
            _value = self.apply_uom(_value, uom)
            _value = self._dtype.encode(_value)
            return _value.encode(encoding) if encoding else _value
        except (ValueError, TypeError) as exc:
            raise ValueError(
                "Output encoding error: '%s' (value '%s')" % (str(exc), value)
            )

    def parse(self, raw_value, uom=None, encoding="utf-8"):
        """ Parse the input value from its string representation.

            The value is checked to match the defined allowed values
            restriction and the UOM conversion is applied.

            Non-Unicode raw_data are converted to Unicode before parsing.
            Byte strings are decoded using the profited encoding (utf8 by
            default).
        """

        try:
            if isinstance(raw_value, text_type):
                _value = raw_value
            elif isinstance(raw_value, str):
                _value = smart_text(raw_value, encoding)
            else:
                _value = smart_text(raw_value)
            _value = self._dtype.parse(raw_value)
            _value = self.strip_uom(_value, uom or self.default_uom)
            _value = self._allowed_values.verify(_value)

            return _value
        except (ValueError, TypeError) as exc:

            raise ValueError(
                "Input parsing error: '%s' (raw value '%s')" % (exc, raw_value)
            )



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2016 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
# pylint: disable=missing-docstring

from sys import stdout
import json
from optparse import make_option
try:
    # available in Python 2.7+
    from collections import OrderedDict
except ImportError:
    from django.utils.datastructures import SortedDict as OrderedDict
from django.utils.six import string_types
from django.core.management.base import BaseCommand, CommandError
from eoxserver.resources.coverages.management.commands import CommandOutputMixIn
from eoxserver.services.models import WMSRenderOptions

JSON_OPTIONS = {
    "indent": None,
    "separators": (', ', ': '),
    "sort_keys": False,
}

class Command(CommandOutputMixIn, BaseCommand):

    option_list = BaseCommand.option_list + (
        make_option(
            '--json', dest='json_dump', action='store_true', default=False,
            help=(
                "Optional. Dump rangetype(s) in JSON format. This JSON "
                "dump can be loaded by another instance of EOxServer."
            )
        ),
        make_option(
            '-o', '--output', dest='filename', action='store', type='string',
            default='-', help=(
                "Optional. Write output to a file rather than to the default"
                " standard output."
            )
        ),
    )

    args = "[<eo-id> [<eo-id> ...]]"

    help = """
    Print WMS options for listed coverages or all coverages. If the coverage
    does not exist or if it exists but it has WMS option record no options
    are printed.
    """

    def handle(self, *args, **options):
        # collect input parameters
        self.verbosity = int(options.get('verbosity', 1))
        print_json = bool(options.get('json_dump', False))
        filename = options.get('filename', '-')

        # get the range types
        wms_opts = WMSRenderOptions.objects.select_related('coverage')
        if args:
            wms_opts = wms_opts.filter(coverage__identifier__in=args)
        else:
            wms_opts = wms_opts.all()

        # select the right output formatter
        if print_json:
            output_formatter = output_json
        else:
            output_formatter = output_detailed

        # write the output
        try:
            with (stdout if filename == "-" else open(filename, "w")) as fout:
                for item in output_formatter(wms_opts):
                    fout.write(item)
        except IOError as exc:
            raise CommandError(
                "Failed to write the output file %r! %s" % (filename, str(exc))
            )


def output_detailed(wms_opts):
    """ Detailed range-type output (includes brief bands' info). """
    for wms_opt in wms_opts:
        yield "%s\n" % wms_opt.coverage.identifier
        for key, val in wms_opt_to_dict(wms_opt).items():
            if hasattr(val, '__len__') and not isinstance(val, string_types):
                val = ",".join(str(v) for v in val)
            if key == "coverage_identifier":
                continue
            if key == "scale_auto" and not val:
                continue
            yield "    %s: %s\n" % (key, val)
        yield "\n"


def output_json(wms_opts):
    """ Full JSON range-type dump. """
    yield '['
    delimiter = ''
    for wms_opt in wms_opts:
        yield delimiter
        yield json.dumps(wms_opt_to_dict(wms_opt), **JSON_OPTIONS)
        delimiter = ',\n'
    yield ']\n'


def wms_opt_to_dict(wms_opt):
    """ Dump objects object to a JSON serializable dictionary. """
    options = OrderedDict()
    options['coverage_identifier'] = wms_opt.coverage.identifier
    if wms_opt.default_red is not None:
        options['red'] = wms_opt.default_red
    if wms_opt.default_green is not None:
        options['green'] = wms_opt.default_green
    if wms_opt.default_blue is not None:
        options['blue'] = wms_opt.default_blue
    if wms_opt.default_alpha is not None:
        options['alpha'] = wms_opt.default_alpha
    if wms_opt.resampling is not None:
        options['resampling'] = wms_opt.resampling
    if wms_opt.scale_auto:
        options['scale_auto'] = wms_opt.scale_auto
    if wms_opt.scale_min is not None:
        options['scale_min'] = wms_opt.scale_min
    if wms_opt.scale_max is not None:
        options['scale_max'] = wms_opt.scale_max
    if wms_opt.bands_scale_min is not None:
        options['bands_scale_min'] = [
            int(v) for v in wms_opt.bands_scale_min.split(',')
        ]
    if wms_opt.bands_scale_max is not None:
        options['bands_scale_max'] = [
            int(v) for v in wms_opt.bands_scale_max.split(',')
        ]
    return options
